""" @brief  A module for solving frequency pairs yielded by the quantum part of
            Shor's algorithm for the general discrete logarithm problem
            [Shor94], modified as in [E19p] and [E24t] (see Sect. 5.5), for the
            logarithm given the order.

    [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                          Logarithms and Factoring".
                         In: Proceedings from FOCS '94, pp. 124–134 (1994).

    [E19p] Ekerå, M.: "Revisiting Shor's quantum algorithm for computing general
                       discrete logarithms".
                      ArXiv 1905.09084v4 (2024).

    [E24t] Ekerå, M.: "On factoring integers, and computing discrete logarithms
                       and orders, quantumly". PhD thesis, KTH Royal Institute
                      of Technology (2024). """

from enum import Enum;

from math import gcd;

from gmpy2 import mpz;

from gmpy2 import invert as mpz_inv;

from ....math.groups import CyclicGroupElement;

from ....math.modular import truncmod;

from ....math.lattices.babai import babai;

from ....math.lattices.cvp import solve_cvp;

from ....math.lattices.lll import lll;
from ....math.lattices.lll import LLL_DEFAULT_DELTA;
from ....math.lattices.lll import LLL_DEFAULT_REDUCED_PRECISION;

from ....math.lattices.enumerate import enumerate as enumerate_lattice;

from ....utils.timeout import Timeout;


B_DEFAULT_ETA = 1000;

B_DEFAULT_T = 1000;


class EnumerationOptions(Enum):

  """ @brief  An enumeration of options for solving a list of n frequency pairs
              [[j_1, k_1], ..., [j_n, k_n]] for d given r.

      The meaning of the options are explained in the documentation for the
      solve_multiple_j_k_for_d_given_r() function. """

  TRUE = True;

  FALSE = False;

  CVP = "CVP";

  BOUNDED_BY_TAU = "BOUNDED_BY_TAU"


def solve_j_k_for_d_given_r(
  j : int | mpz,
  k : int | mpz,
  m : int,
  sigma : int,
  l : int,
  g : CyclicGroupElement,
  x : CyclicGroupElement,
  r : int | mpz,
  B_ETA : int = B_DEFAULT_ETA,
  B_T : int = B_DEFAULT_T,
  timeout : int | None | Timeout = None,
  verbose : bool = False) -> int | mpz | None:

  """ @brief  Attempts to compute the general discrete logarithm d = log_g x
              given a frequency pair (j, k) yielded by the quantum part of
              Shor's algorithm [Shor94], modified as in [E19p], and the order r
              of g, by using the post-processing algorithm from [E19p].

      More specifically, the post-processing algorithm is in Sect. 6 of [E19p].

      [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                           In: Proceedings from FOCS '94, pp. 124–134 (1994).

      [E19p] Ekerå, M.: "Revisiting Shor's quantum algorithm for computing
                         general discrete logarithms".
                        ArXiv 1905.09084v4 (2024).

      Note that this function does not implement meet-in-the-middle-techniques,
      although it is noted in [E23p] that it is possible to use such techniques
      to speed up the post-processing also for general discrete logarithms.

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025).

      @param j  The frequency j. An integer on [0, 2^(m + sigma)).

      @param k  The frequency k. An integer on [0, 2^l).

      @param m  A positive integer m such that r < 2^m.

      @param sigma  A non-negative integer sigma such that m + sigma is the
                    length of the first control register in the quantum part of
                    the algorithm.

      @param l  A positive integer l such that l is the length of the second
                control register in the quantum part of the algorithm.

      @param g  The group element g.

      @param x  The group element x = g^d.

      @param r  The order r of g.

      @param B_ETA  An upper bound on the search space in eta.

      @param B_T  An upper bound on the search space in t.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.


      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when executing the
                      post-processing algorithm.

      @return   The logarithm d, or None, if solving for d fails. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  j = mpz(j);
  k = mpz(k);

  r = mpz(r);

  # Sanity checks.
  if (m <= 0) or (not (r < (2 ** m))):
    raise Exception("Error: Incorrect parameter m.");

  if l <= 0:
    raise Exception("Error: Incorrect parameter l.");

  if sigma < 0:
    raise Exception("Error: Incorrect parameter sigma.");

  # Pre-compute z and w.
  z = (r * j - truncmod(mpz(r * j), mpz(2 ** (m + sigma)))) // \
    mpz(2 ** (m + sigma));
  w = (r * k - truncmod(mpz(r * k), mpz(2 ** l))) // mpz(2 ** l);

  for abs_eta in range(B_ETA):

    # Check the timeout.
    timeout.check();

    for sgn_eta in [1, -1]:
      if (abs_eta == 0) and (sgn_eta == -1):
        continue;
      eta = sgn_eta * abs_eta;

      if gcd(mpz(z + eta), mpz(r)) != 1:
        continue;

      # Precompute (z + eta)^-1 (mod r).
      z_plus_eta_inv = mpz_inv(mpz(z + eta), mpz(r));

      # Recall that: candidate_d = (t - w) * z_plus_eta_inv
      #
      # As t increases or decreases, we step d by w * z_plus_eta_inv or by
      # -w * z_plus_eta_inv, respectively.
      #
      # Let us pre-compute g to the power of these two exponents:
      x_step_plus  = g ** z_plus_eta_inv;
      x_step_minus = x_step_plus ** -1; # = g ** (-z_plus_eta_inv);

      # The starting point x0 is -w * z_plus_eta_inv.
      #
      # Again, let us pre-compute g to the power of this exponent:
      x0 = x_step_minus ** w; # = g ** (-w * z_plus_eta_inv)

      # Since we search from x0 out, setup storage for intermediary values:
      [candidate_x_plus, candidate_x_minus] = [x0, x0];

      for abs_t in range(B_T):

        # Check the timeout.
        timeout.check();

        for sgn_t in [1, -1]:
          if (abs_t == 0) and (sgn_t == -1):
            continue;
          t = sgn_t * abs_t;

          # Compute candidate_d:
          candidate_d = ((t - w) * z_plus_eta_inv) % r;

          # Set candidate_x = g^candidate_d by using the pre-computed values:
          if sgn_t == 1:
            candidate_x = candidate_x_plus;
            candidate_x_plus = candidate_x_plus * x_step_plus;
          else:
            candidate_x_minus = candidate_x_minus * x_step_minus;
            candidate_x = candidate_x_minus;

          if verbose:
            print("Candidate d:", candidate_d, "t:", t, "eta:", eta);

          if candidate_x == x:
            return candidate_d;

  # Failed to solve for d.
  return None;


def solve_multiple_j_k_for_d_given_r(
  j_k_list : list[list[int | mpz]],
  m : int,
  sigma : int,
  l : int,
  g : CyclicGroupElement,
  x : CyclicGroupElement,
  r : int | mpz,
  tau : int = 0,
  delta : float = LLL_DEFAULT_DELTA,
  precision : int | None = None,
  enumerate : bool | EnumerationOptions = False,
  timeout : int | None | Timeout = None,
  verbose : bool = False) -> int | mpz | None:

  """ @brief  Attempts to compute the general discrete logarithm d = log_g x
              given a list of n frequency pairs [[j_1, k_1], ..., [j_n, k_n]]
              yielded by n independent runs of the quantum part of Shor's
              algorithm [Shor94] for computing general discrete logarithms,
              modified as in [E19p], and the order r of g, by using the
              lattice-based post-processing algorithm described in [E19p]
              (see Sect. 6) and [E24t] (see Sect. 5.5).

      [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                           In: Proceedings from FOCS '94, pp. 124–134 (1994).

      [E19p] Ekerå, M.: "Revisiting Shor's quantum algorithm for computing
                         general discrete logarithms".
                        ArXiv 1905.09084v4 (2024).

      [E24t] Ekerå, M.: "On factoring integers, and computing discrete
                         logarithms and orders, quantumly".
                        PhD thesis, KTH Royal Institute of Technology (2024).

      Note that this function does not implement meet-in-the-middle-techniques,
      although it is noted in [E23p] that it is possible to use such techniques
      to speed up the post-processing also for general discrete logarithms.

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025).

      @param j_k_list   The n frequency pairs [[j_1, k_1], ..., [j_n, k_n]]
                        where j_1, ..., j_n are integers on [0, 2^(m + sigma))
                        and k_1, ..., k_n are integers on [0, 2^l).

      @param m  A positive integer m such that r < 2^m.

      @param sigma  A non-negative integer sigma such that m + sigma is the
                    length of the first control register in the quantum part
                    of the algorithm.

      @param l  A positive integer l ≈ m / s for s a tradeoff factor such that
                l is the length of the second control register in the quantum
                part of the algorithm.

      @param g  The group element g.

      @param x  The group element x = g^d.

      @param r  The order r of g.

      @param tau  A positive integer tau. Used to scale the basis for the
                  lattice L^tau that is used in the post-processing.

      @param delta  The parameter delta to use when delta-LLL-reducing the basis
                    for the lattice L^tau used in the post-processing. Must be
                    on the interval (1/4, 1]. A polynomial runtime in the
                    dimension of the lattice is only guaranteed for delta < 1.

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors as a part of delta-LLL-reducing the
                        basis for the lattice L^tau used in the post-processing.

                        The precision may be set to None, as is the default, in
                        which case the projection factors are represented as
                        exact quotients.

      @param enumerate  A flag that may be set to True to enumerate vectors in
                        the lattice L^tau (until d is found or the specified
                        timeout has elapsed), or to EnumerationOptions.CVP to
                        consider only a closest vector in the lattice as
                        returned by performing a limited enumeration, or to
                        False to consider only the vector returned by Babai's
                        nearest plane algorithm.

                        May also be set to EnumerationOptions.BOUNDED_BY_TAU in
                        which case all vectors within distance R of the origin
                        of the lattice L^tau are enumerated, where R depends on
                        tau as R = sqrt(n + 1) * 2^(m + sigma - l + tau).

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when executing the
                      post-processing algorithm.

      @return   The logarithm d, or None, if solving for d fails. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  if None in j_k_list:
    return None;

  n = len(j_k_list);
  if n < 1:
    return None;

  js, ks = zip(*j_k_list)

  # Setup a basis for the lattice.
  if verbose:
    print("Setting up the basis...");

  row = [];

  for i in range(n):
    alpha_r_i = mpz(truncmod(r * js[i], 2 ** (m + sigma)));
    row.append((mpz(r * js[i]) - alpha_r_i) * mpz(2 ** l));

  row.append(mpz(2 ** (sigma + tau)) * mpz(r));

  A = [row];

  for i in range(n):
    row = (n + 1) * [mpz(0)];
    row[i] = mpz(2 ** (m + sigma + l)) * mpz(r);
    A.append(row);

  # Reduce the basis for the lattice.
  if verbose:
    print("Computing the LLL-reduced basis with reduced precision...");

  B = lll(A,
          delta = delta,
          timeout = timeout,
          precision = LLL_DEFAULT_REDUCED_PRECISION);
            # Pre-compute with reduced precision.

  # Check the timeout.
  timeout.check();

  if verbose:
    print(" ** Refining the LLL-reduced basis with full precision...");

  [B, [Bs, M]] = lll(B,
                     delta = delta,
                     timeout = timeout,
                     gs = True,
                     precision = precision);

  # Check the timeout.
  timeout.check();

  # Setup v.
  if verbose:
    print("Setting up the target vector v...");

  v = [-(mpz(2 ** (m + sigma)) * mpz(k) * mpz(r)) for k in ks] + [mpz(0)];

  if enumerate in [EnumerationOptions.TRUE, True,
                   EnumerationOptions.FALSE, False,
                   EnumerationOptions.CVP]:

    # Find a vector u in the lattice that is close to v.
    if enumerate == EnumerationOptions.CVP:
      if verbose:
        print("Executing an enumeration to solve CVP...");

      # Let u be a closest vector to v in the lattice.
      u = solve_cvp(B, v, timeout = timeout, gs = [Bs, M]);
    else:
      if verbose:
        print("Executing Babai's nearest plane algorithm...");

      # Let u be a close vector to v in the lattice.
      u = babai(B, v, Bs);

    # Check if the correct solution was found.
    if verbose:
      print(" ** Checking:", u);

    candidate_d = u[-1] // mpz(2 ** (sigma + tau)) // mpz(r);

    if -r < candidate_d < 0:
      # Probably d - r was recovered instead of d.
      candidate_d += r;

    if 0 <= candidate_d < r:
      if g ** candidate_d == x:
        return candidate_d;

    if enumerate not in [True, EnumerationOptions.TRUE]:
      return None;

    # Check the timeout.
    timeout.check();

    if verbose:
      print("Enumerating vectors around v...");

    # Set the enumeration radius R = |u - v|^2.
    R2 = sum([(u[i] - v[i]) ** 2 for i in range(n + 1)]);

    # Enumerate with gradually increasing radii.
    while True:

      # Check the timeout.
      timeout.check();

      # Enumerate, whilst checking the timeout.
      for u in enumerate_lattice(B,
                                 R2,
                                 v,
                                 timeout = timeout,
                                 gs = [Bs, M]):

        # Check if the correct solution was found.
        if verbose:
          print(" ** Checking:", u);

        candidate_d = u[-1] // mpz(2 ** (sigma + tau)) // mpz(r);

        if -r < candidate_d < 0:
          # Probably d - r was recovered instead of d.
          candidate_d += r;

        if 0 <= candidate_d < r:
          if g ** candidate_d == x:
            return candidate_d;

      # Double the enumeration radius.
      if verbose:
          print(" ** Doubling the enumeration radius.");

      R2 = 4 * R2;

  elif enumerate == EnumerationOptions.BOUNDED_BY_TAU:

    # Set the enumeration radius R.
    R2 = mpz(n + 1) * mpz(2 ** (2 * (m + sigma + tau))) * mpz(r * r);

    # Enumerate, whilst checking the timeout.
    for u in enumerate_lattice(B,
                               R2,
                               v,
                               timeout = timeout,
                               gs = [Bs, M]):

      # Check if the correct solution was found.
      if verbose:
        print(" ** Checking:", u);

      candidate_d = u[-1] // mpz(2 ** (sigma + tau)) // mpz(r);

      if -r < candidate_d < 0:
        # Probably d - r was recovered instead of d.
        candidate_d += r;

      if 0 <= candidate_d < r:
        if g ** candidate_d == x:
          return candidate_d;

    # No solution was found.
    return None;

  else:
    raise Exception("Error: Incorrect parameters: Unknown enumerate option.");


""" @brief  A module for a utility class for handling timeouts. """

from __future__ import annotations;

from .timer import Timer;

class Timeout:

  """ @brief  A utility class for handling timeouts. """

  def __init__(self, timeout : int | None = None):

    """ @brief  Initializes the timeout to a specific time in seconds.

        @param timeout  The timeout in seconds. May be set to None in which case
                        no timeout is enforced. """

    self.timer = Timer().start();
    self.timeout = timeout;

  def check(self) -> None:

    """ @brief  Raises a TimeoutError if the timeout is elapsed. """

    if self.is_elapsed():
      raise TimeoutError("Error: The timeout has elapsed.");

  def is_elapsed(self) -> bool:

    """ @brief  Returns True if the timeout is elapsed, False otherwise.

        @return   True if the timeout is elapsed, False otherwise. """

    if None == self.timeout:
      return False;

    return self.timer.peek() > self.timeout;

  def is_indefinite(self) -> bool:

    """ @brief  Returns True if the timeout in indefinite, False otherwise.

        @return   True if the timeout is indefinite, False otherwise. """

    return self.timeout == None;

  def __str__(self) -> str:

    """ @brief  Returns a string representation of the timeout.

        @return   A string representation of the timeout. """

    if None == self.timeout:
      return "An indefinite timeout.";
    else:
      s = Timer.format(self.timer.peek());

      if not self.is_elapsed():
        return "A timeout of " + Timer.format(self.timeout) + " with " + s + \
          " elapsed.";
      else:
        return "A timeout of " + Timer.format(self.timeout) + " that is " + \
          "elapsed.";

  def __repr__(self) -> str:

    """ @brief  Returns a string representation of the timeout.

        @return   A string representation of the timeout. """

    return str(self);

  @staticmethod
  def parse(timeout : int | None | Timeout) -> Timeout:

    """ @brief  Parses a timeout that may be an integer, None or an instance of
                Timeout and returns a corresponding instance of Timeout.

        This is a static convenience function.

        @param timeout  The timeout to parse.

        @return   An instance of Timeout representing the timeout. """

    if isinstance(timeout, Timeout):
      return timeout;

    if None == timeout:
      return Timeout(None);

    if isinstance(timeout, int):
      if timeout < 0:
        raise ValueError("Error: Incorrect timeout.");

      return Timeout(timeout);

    raise ValueError("Error: Incorrect timeout.");


""" @brief  A module for splitting N, given the order r of an element g selected
            uniformly at random from the multiplicative group of the ring of
            integers modulo N, where g must be explicitly specified. This by
            using the original algorithm from [Shor94].

    [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                          Logarithms and Factoring".
                         In: Proceedings from FOCS '94, pp. 124–134 (1994). """

from gmpy2 import mpz;

from gmpy2 import gcd;
from gmpy2 import powmod;

def split_N_given_g_r(
  g : int | mpz,
  r : int | mpz,
  N : int | mpz) -> set[int] | None:

  """ @brief  Attempts to split N, given the order r of an element g selected
              uniformly at random from the multiplicative group of the ring of
              integers modulo N, where g must be explicitly specified. This by
              using the original algorithm from [Shor94].

      [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                           In: Proceedings from FOCS '94, pp. 124–134 (1994).

      The algorithm succeeds iff r is even and g^(r/2) != -1 (mod N).

      @param g  The element g.

      @param r  The order r of g mod N.

      @param N  The modulus N. As in [Shor94], it is required that N is odd,
                and not a perfect prime power.

      @return   A set {p, q} of two non-trivial factors of N such that N = pq,
                or None if the algorithm failed to split N. """

  # Convert inputs.
  g = mpz(g);
  r = mpz(r);
  N = mpz(N);

  # Sanity checks.
  if not (0 < g < N):
    raise Exception("Error: Incorrect parameters: g is not on [1, N).");

  if (N % 2) == 0:
    raise Exception("Error: Incorrect parameters: N is not odd.");

  if not (0 < r < N / 2):
    raise Exception("Error: Incorrect parameters: N is not on [1, N/2).");

  if not gcd(g, N) == 1:
    raise Exception("Error: Incorrect parameters: g is not coprime to N.");

  # Attempt to split N.
  if (r % 2) != 0:
    # Additional sanity check.
    if powmod(g, r, N) != 1:
      raise Exception("Error: Incorrect parameters: r is not the order of g.");

    return None;

  x = powmod(g, r // 2, N);

  # Additional sanity check.
  if (x == 1) or (powmod(x, 2, N) != 1):
    raise Exception("Error: Incorrect parameters: r is not the order of g.");

  if x == N - 1:
    return None;

  [p, q] = [gcd(x - 1, N), gcd(x + 1, N)];

  # Additional sanity check.
  if N != p * q:
    raise Exception("Error: Internal error: Is N a perfect prime power?");

  return set([int(p), int(q)]);


""" @brief  A module for splitting N = pq into the large l-bit prime factors
            p and q given d = (p - 1) / 2 + (q - 1) / 2 - 2^(l - 1). """

from gmpy2 import mpz;

from gmpy2 import isqrt_rem as sqrt_rem;
from gmpy2 import ceil;

def split_N_given_d(
  d : int | mpz,
  N : int | mpz) -> set[int] | None:

  """ @brief  Splits N = pq into the large l-bit prime factors p and q given
              d = (p - 1) / 2 + (q - 1) / 2 - 2^(l - 1).

      This is a convenience function for Ekerå–Håstad's algorithm [EH17] that
      factors a large random RSA integer N = pq into p and q.

      More specifically, as is explained in App. A.2 of [E20], it holds that

        x = g^d' = g^((N - 1) / 2 - 2^(l - 1))
                 = g^((p - 1) / 2 + (q - 1) / 2 - 2^(l - 1)) = g^d

      Provided that the order r of g is sufficiently large, it holds that

        d = d' mod r = (p - 1) / 2 + (q - 1) / 2 - 2^(l - 1),

      allowing d to be computed as the discrete logarithm of x to the base g.

      This may be done efficiently by using Ekerå–Håstad's quantum algorithm
      that computes short discrete logarithms in groups of unknown order.

      Given d = (p - 1) / 2 + (q - 1) / 2 - 2^(l - 1) and N = pq, it is then
      trivial to compute p and q by solving a quadratic equation.

      This convenience function performs the last step in the above procedure.

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E20] Ekerå, M.: "On post-processing in the quantum algorithm for
                        computing short discrete logarithms".
                       Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

      @param d  The logarithm d.

      @param N  The integer N. It is required that N is odd, and the product
                of two large random distinct l-bit prime numbers.

      @return   A set {p, q} of two non-trivial factors of N such that N = pq,
                or None if the algorithm failed to split N. """

  # Convert inputs.
  d = mpz(d);
  N = mpz(N);

  # Sanity checks.
  if (N % 2) == 0:
    raise Exception("Error: Incorrect parameters: N is not odd.");

  if not (0 < d < N / 2):
    raise Exception("Error: Incorrect parameters: d is not on [1, N/2).");

  # Compute the bit length l of p and q where N = pq: We have that
  #
  #   p, q >= 2^(l-1) => N = pq = 2^(2(l-1))
  #                        = 2^(2l-2),
  #
  # i.e. N is an n >= 2l - 1 bit integer. At the same time, we have that
  #
  #   p, q < 2^l => N = pq < 2^2l,
  #
  # i.e. N is an n <= 2l bit integer. Hence, for n the bit length of N, it must
  # be that l = ceil(n / 2), so we may compute l from n in this manner:
  l = int(ceil(N.bit_length() / 2));

  # Form d' = p + q given d.
  p_plus_q = 2 * d + 2 ** l + 2;

  # Solve d' = p + q and N = pq for p and q using the quadratic formula:
  if ((p_plus_q ** 2) - 4 * N <= 0) or (0 != (p_plus_q % 2)):
    # Failed to factor.
    return None;

  (root, remainder) = sqrt_rem((p_plus_q ** 2) - 4 * N);

  # Sanity checks.
  if (0 != remainder) or (0 != (root % 2)):
    # Failed to factor.
    return None;

  # Compute p and q.
  p = (p_plus_q - root) // 2;
  q = (p_plus_q + root) // 2;

  # Sanity checks.
  if (1 < p < N) and (1 < q < N) and (p * q == N):

    # Factored N into p and q.
    return set([int(p), int(q)]);

  # Failed to factor.
  return None;


""" @brief  A module for sampling random integers. """

from secrets import randbelow;

def sample_integer(B : int) -> int:

  """ @brief  Returns an integer selected uniformly at random from [0, B).

      @remark   This function calls randbelow() to select an integer uniformly
                at random from [0, B). In practice, randbelow() returns an
                integer that may be conjectured to be indistinguishable from an
                integer that is selected uniformly at random from [0, B).

      @param B  The upper bound B.

      @return   An integer selected uniformly at random from [0, B). """

  if B < 1:
    raise Exception("Error: Incorrect parameters: Pick B >= 1.");

  return randbelow(B);


def sample_l_bit_integer(l : int) -> int:

  """ @brief  Returns an l-bit integer selected uniformly at random from the set
              of all such integers.

      @remark   This function calls sample_integer() to select an integer
                uniformly at random from [0, B) for B = 2^(l-1). In practice,
                sample_integer() returns an integer that may be conjectured to
                be indistinguishable from an integer that is selected uniformly
                at random from [0, B).

      @param l  The bit length l of the prime to sample.

      @return   An l-bit integer selected uniformly at random from the set of
                all such integers. """

  if l < 1:
    raise Exception("Error: Incorrect parameters: Pick l >= 1.");

  return 2 ** (l - 1) + sample_integer(2 ** (l - 1));


""" @brief  A module for finding the solution to a set of congruence relations
            with coprime moduli via the Chinese remainder theorem. """

from gmpy2 import mpz;

from gmpy2 import invert;

from math import prod;

def crt(
    values : list[int | mpz],
    moduli : list[int | mpz]) -> int | mpz:

  """ @brief  Given values = [v1, ..., vn] and moduli = [N1, ..., Nn], this
              function returns an integer v in [0, N) such that v = vi (mod Ni)
              for all i in [1, n], where N = N1 * ... * Nn.

      @remark It is required that all Ni >= 2, and pairwise coprime.

      @param values   The values = [v1, ..., vn].

      @param moduli   The moduli = [N1, ..., Nn].

      @return   An integer v in [0, N) such that v = vi (mod Ni) for all i in
                [1, n], where N = N1 * ... * Nn. """

  n = len(values);

  if (n != len(moduli)) or (n < 1):
    raise Exception("Error: Incorrect parameters.");

  N = prod([Ni for Ni in moduli]);

  Tis = [N // Ni for Ni in moduli];

  Tinvis = [invert(Tis[i], moduli[i]) for i in range(n)];

  return sum([values[i] * Tinvis[i] * Tis[i] for i in range(n)]) % N;


""" @brief  A module for Gram–Schmidt orthogonalization. """

import gmpy2;

from gmpy2 import mpz;
from gmpy2 import mpq;
from gmpy2 import mpfr;

from ..matrices import dimensions;

from copy import deepcopy;

def gram_schmidt(
  B : list[list[int | mpz]],
  precision : int | None = None) -> \
    list[list[list[int | mpz | mpq | mpfr]]]:

  """ @brief  Returns the Gram–Schmidt orthogonalization Bs of an n x d basis
              matrix B, and the associated matrix M of Gram–Schmidt projection
              factors which is such that B = M Bs.

      The integer matrix B = [b_1, ..., b_n], where b_i = [b_i1, ..., b_id] for
      i = 1, ..., n are the n row vectors of B. The matrices Bs and M are
      represented in analogy with how B is represented.

      @param B  The matrix B.

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors in M. May be set to None, as is the
                        default, in which case the projection factors are
                        represented as exact quotients.

      @return The pair [Bs, M], where Bs is the Gram–Schmidt orthogonalization
              of B and M is the matrix of Gram–Schmidt projection factors. """

  (n, d) = dimensions(B);

  B = [[mpz(x) for x in row] for row in B];

  Bs = deepcopy(B);

  M = [[mpz(0) for _ in range(d)] for _ in range(n)];
  for i in range(n):
    M[i][i] = mpz(1);

  with gmpy2.context(gmpy2.get_context()) as context:

    # Set the precision.
    if precision != None:
      context.precision = precision;

    def compute_mu(i, j):
      a = sum([B[i][k]  * Bs[j][k] for k in range(d)]);
      b = sum([Bs[j][k] * Bs[j][k] for k in range(d)]);

      if None == precision:
        return mpq(a, b);

      with gmpy2.context(gmpy2.get_context()) as context:

        # Set the precision.
        context.precision = precision;

        return mpfr(a) / mpfr(b);

    for i in range(2, n+1):
      for j in range(i-1, 0, -1):
        M[i-1][j-1] = compute_mu(i-1, j-1);

        for k in range(d):
          Bs[i-1][k] = Bs[i-1][k] - M[i-1][j-1] * Bs[j-1][k];

    return [Bs, M];


""" @brief  A module for performing or simulating arithmetic operations in
            cyclic groups. """

from abc import abstractmethod;

from gmpy2 import mpz;

from gmpy2 import gcd;
from gmpy2 import powmod;

from gmpy2 import invert as mpz_inv;

class CyclicGroupElement:

  """ @brief  An abstract class that represents a group element in a finite
              cyclic group, that is written multiplicatively. """

  @abstractmethod
  def __pow__(self, e : int):

    """ @brief  Returns the group element g^e, for g this group element.

        @param e  The exponent e.

        @return   The group element g^e, for g this group element. """

    pass;

  @abstractmethod
  def __mul__(self, x):

    """ @brief  Returns the group element g * x, for g this group element.

        @param x  The group element x.

        @return   The group element g * x, for g this group element. """

    pass;

  @abstractmethod
  def __eq__(self, x):

    """ @brief  Compares this group element to another group element.

        The two elements must be in the same group.

        @param x  The group element to which to compare this group element, or
                  one (1) to compare this group element to the identity element.

        @return   True if x is equal to this group element, False otherwise. """

    pass;

  @abstractmethod
  def __hash__(self):

    """ @brief  Returns the hash digest of this group element.

        @return   The hash digest of this group element. """

    pass;


class IntegerModRingMulSubgroupElement(CyclicGroupElement):

  """ @brief  A class that represents the group element g in Z_N^*. """

  def __init__(self, g, N):

    """ @brief  Constructs the group element g in Z_N^*.

        @param g  The element g. An integer in [0, N), coprime to N.

        @param N  The modulus N. A positive integer. """

    self.g = mpz(g);
    self.N = mpz(N);

    if not ((1 <= self.g < self.N) and (1 == gcd(self.g, self.N))):
      raise Exception("Error: Incorrect parameters.");

  def __pow__(self, e : int):

    """ @brief  Returns the group element g^e, for g this group element.

        @param e  The exponent e.

        @return   The group element g^e, for g this group element. """

    if e < 0:
      x = mpz_inv(self.g, self.N);
      x = powmod(x, mpz(-e), self.N);
    else:
      x = powmod(self.g, mpz(e), self.N);

    return IntegerModRingMulSubgroupElement(x, self.N);

  def __mul__(self, x):

    """ @brief  Returns the group element g * x, for g this group element.

        @param x  The group element x.

        @return   The group element g * x, for g this group element. """

    return IntegerModRingMulSubgroupElement((self.g * x.g) % self.N, self.N);

  def __eq__(self, x):

    """ @brief  Compares this group element to another group element.

        @param x  The group element to which to compare this group element, or
                  one (1) to compare this group element to the identity element.

        @return   True if x is equal to this group element, False otherwise. """

    if (x == 1) and (self.g == 1):
      return True;

    if type(x) != type(self):
      return False;

    if x.N != self.N:
      return False;

    return self.g == x.g;

  def __hash__(self):

    """ @brief  Returns the hash digest of this group element.

        @return   The hash digest of this group element. """

    return hash((self.g, self.N));

  def __str__(self):

    """ @brief  Returns a string representation of the group element.

        @return   A string representation of the group element. """

    return str(self.g) + " (mod " + str(self.N) + ")";

  def __repr__(self):

    """ @brief  Returns a string representation of the group element.

        @return   A string representation of the group element. """

    return str(self);


class ShortWeierstrassCurveOverPrimeField:

  """ @brief  A class that represents an elliptic curve on short Weierstrass
              form y^2 = x^3 + ax + b (mod p) for p a prime number. """

  def __init__(self, a, b, p):

    """ @brief  Constructs an elliptic curve on short Weierstrass form
                y^2 = x^3 + ax + b (mod p) for p a prime number.

        @param a  The Weierstrass coefficient a.

        @param b  The Weierstrass coefficient b.

        @param p  The prime p. """

    self.p = mpz(p);

    self.a = mpz(a) % self.p;
    self.b = mpz(b) % self.p;

  def identity_y(self):

    """ @brief  Returns the y coordinate of the point used to represent the
                point at infinity on this curve.

        The point at infinity is represented by (0, self.identity_y()), where
        the y coordinate is 0 if (0, 0) is not in E, and (0, 1) otherwise.

        @return   The y coordinate of the point used to represent the
                  point at infinity on this curve."""

    if 0 == self.b:
      return 1; # Point at infinity is represented as (0, 1).
    else:
      return 0; # Point at infinity is represented as (0, 0).

  def __eq__(self, E):

    """ @brief  Compares this curve to another curve.

        @param E  The curve to which to compare this curve.

        @return   True if E is equal to this curve, False otherwise. """

    return (self.p == E.p) and (self.a == E.a) and (self.b == E.b);

  def __str__(self):

    """ @brief  Returns a string representation of the curve.

        @return   A string representation of the curve. """

    return "y^2 = x^3 + " + str(self.a) + " x + " + str(self.b) + \
      " (mod " + str(self.p) + ")";

  def __repr__(self):

    """ @brief  Returns a string representation of the curve.

        @return   A string representation of the curve. """

    return str(self);


class PointOnShortWeierstrassCurveOverPrimeField(CyclicGroupElement):

  """ @brief  A class that represents a point (x, y) on a short Weierstrass
              curve over a prime field.

      Note that we write elliptic curve groups multiplicatively: This is
      non-standard, but beneficial since it allows elliptic curve groups to
      be substituted for multiplicative groups, and vice versa, in function
      calls. More specifically, P + Q, [e] P and [-1] P becomes P * Q, P^e
      and P^-1, respectively, when writing the group multiplicatively. """

  def __init__(self, x, y, E):

    """ @brief  Constructs the point (x, y) on the elliptic curve E.

        @param x  The x coordinate.

        @param y  The y coordinate.

        @param E  The elliptic curve E. """

    self.x = mpz(x) % E.p;
    self.y = mpz(y) % E.p;
    self.E = E;

    if not self.is_identity():
      if 0 != (((self.x ** 3) + E.a * self.x + E.b - (self.y ** 2)) % E.p):
        raise Exception("Error: The point is not on the curve.");

  def is_identity(self):

    """ @brief  Returns True if this point is the identity, False otherwise.

        @return   True if this point is the identity, False otherwise. """

    return (0 == self.x) and (self.E.identity_y() == self.y);

  def invert(self):

    """ @brief  Returns the point P^-1, for P this point.

        @return   The point P^-1, for P this point. """

    return PointOnShortWeierstrassCurveOverPrimeField(self.x, -self.y, self.E);

  def __pow__(self, e : int):

    """ @brief  Returns the point P^e, for P this point.

        @param e  The exponent e.

        @return   The point P^e, for P this point. """

    # Set P to this point.
    P = PointOnShortWeierstrassCurveOverPrimeField(self.x, self.y, self.E);

    if e < 0:
      P = P.invert();
      e = -e;

    # Set R to the point at infinity on E.
    R = PointOnShortWeierstrassCurveOverPrimeField(
          0, self.E.identity_y(), self.E);

    while 0 != e:
      if 1 == (e % 2):
        R = R * P;
      P = P * P;
      e = e // 2;

    return R;

  def __mul__(self, Q):

    """ @brief  Returns the point P * Q, for P this point.

        @param Q  The point Q.

        @return   The point P * Q, for P this point. """

    if Q == 1:
      return self;

    if self == 1:
      return Q;

    if self == Q.invert():
      if 0 == self.E.b:
        return PointOnShortWeierstrassCurveOverPrimeField(0, 1, self.E);
      else:
        return PointOnShortWeierstrassCurveOverPrimeField(0, 0, self.E);

    if self.E != Q.E:
      raise Exception("Error: The two points are not on the same curve.");

    if self == Q:
      s = (3 * (self.x ** 2) + self.E.a) * mpz_inv(2 * self.y, self.E.p);
      s = s % self.E.p;
      x = ((s ** 2) - 2 * self.x) % self.E.p;
      y = (-s * (x - self.x) - self.y) % self.E.p;
    else:
      s = ((self.y - Q.y) * mpz_inv(self.x - Q.x, self.E.p)) % self.E.p;
      x = ((s ** 2) - self.x - Q.x) % self.E.p;
      y = ( s * (self.x - x) - self.y) % self.E.p;

    return PointOnShortWeierstrassCurveOverPrimeField(x, y, self.E);

  def __eq__(self, Q):

    """ @brief  Compares this point to another point Q.

        @param Q  The point to which to compare this point, or one (1) to
                  compare this point to the identity element.

        @return   True if Q is equal to this point, False otherwise. """

    if (Q == 1) and (self.x == 0) and (self.y != self.E.b):
      return True;

    if type(Q) != type(self):
      return False;

    if Q.E != self.E:
      return False;

    return (Q.x == self.x) and (Q.y == self.y);

  def __hash__(self):

    """ @brief  Returns the hash digest of this pointgroup element.

        @return   The hash digest of this group element. """

    return hash((self.x, self.y, self.E.p, self.E.a, self.E.b));

  def __str__(self):

    """ @brief  Returns a string representation of the group element.

        @return   A string representation of the group element. """

    if self.is_identity():
      return "Point at infinity on " + str(self.E);

    return "(" + str(self.x) + ", " + str(self.y) + ") on " + str(self.E);

  def __repr__(self):

    """ @brief  Returns a string representation of the group element.

        @return   A string representation of the group element. """

    return str(self);


class SimulatedCyclicGroupElement(CyclicGroupElement):

  """ @brief  A class that represents the simulated group element G^d, for G a
              generator of a cylic group of order r. """

  def __init__(self, r, d = 1):

    """ @brief  Constructs the simulated group element G^d, for G a generator of
                a cylic group of order r.

        @param r  The order r. A positive integer.

        @param d  The exponent d on [0, r). """

    self.r = mpz(r);
    self.d = mpz(d);

    if not (0 <= self.d < self.r):
      raise Exception("Error: Incorrect parameters.");

  def __pow__(self, e : int):

    """ @brief  Returns the group element g^e, for g this group element.

        @param e  The exponent e.

        @return   The group element g^e, for g this group element. """

    return SimulatedCyclicGroupElement(self.r, (self.d * e) % self.r);

  def __mul__(self, x):

    """ @brief  Returns the group element g * x, for g this group element.

        @param x  The group element x.

        @return   The group element g * x, for g this group element. """

    if x == 1:
      return SimulatedCyclicGroupElement(self.r, self.d);

    if type(self) != type(x):
      raise Exception("Error: The two elements must be in the same group.");

    if self.r != x.r:
      raise Exception("Error: The two elements must be in the same group.");

    return SimulatedCyclicGroupElement(self.r, (self.d + x.d) % self.r);

  def __eq__(self, x):

    """ @brief  Compares this group element to another group element.

        @param x  The group element to which to compare this group element, or
                  one to compare this group element to the identity element.

        @return   True if x is equal to this group element, False otherwise. """

    if (x == 1) and (self.d == 0):
      return True;

    if type(x) != type(self):
      return False;

    if x.r != self.r:
      return False;

    return self.d == x.d;

  def __hash__(self):

    """ @brief  Returns the hash digest of this group element.

        @return   The hash digest of this group element. """

    return hash((self.d, self.r));

  def __str__(self):

    """ @brief  Returns a string representation of the group element.

        @return   A string representation of the group element. """

    return "C(" + str(self.d) + " / " + str(self.r) + ")";

  def __repr__(self):

    """ @brief  Returns a string representation of the group element.

        @return   A string representation of the group element. """

    return str(self);


""" @brief  A module for modular arithmetic. """

from gmpy2 import mpz;

def truncmod(x : int | mpz, N : int | mpz) -> int | mpz:

  """ @brief  Returns x mod N constrained to the interval [N/2, N/2).

      @param x  The integer x.

      @param N  The modulus N.

      @return The integer x mod N, constrained to the interval [N/2, N/2). """

  x = x % N;

  if 2 * x >= N:
    x -= N;

  return x;


""" @brief  A module for testing the functions in the parent module for
            factoring N completely.

    This module implements the test procedure described in App. A to [E21b].

    [E21b] Ekerå, M.: "On completely factoring any integer efficiently in a
                       single run of an order-finding algorithm".
                      Quantum Inf. Process. 20(6):205 (2021).

    @remark This module is modelled upon the corresponding Sage script in the
            Factoritall repository, see https://github.com/ekera/factoritall.

    To run all unit tests in this module by calling test_all(), execute:

    ```sh

    $ python3 -m quaspy.factoring.general.postprocessing.ekera.test

    ``` """

from gmpy2 import mpz;

from math import gcd;
from math import prod;
from math import floor;
from math import prod;

from ......utils.timer import Timer;

from ......math.groups import SimulatedCyclicGroupElement;
from ......math.groups import IntegerModRingMulSubgroupElement;

from ......math.random import sample_integer;

from .. import solve_j_for_factors;
from .. import solve_r_for_factors;

from ......orderfinding.general.sampling import sample_j_given_r;

from .....sampling import sample_r_given_N;
from .....sampling import sample_g_r_given_N;

from .....sampling import B_DEFAULT_SAMPLE;

from ......orderfinding.general.postprocessing.ekera import B_DEFAULT_SOLVE;

from ......math.primes import is_B_smooth;

def is_admissible_r(
  r : int | mpz,
  B : int,
  N_factors : list[list[int | mpz]]) -> bool:

  """ @brief  Tests if (pi - 1) / gcd(pi - 1, r) is B-smooth for all but at
              most one of the prime factors pi of N.

      @param r  The order r of an

      @param B  The upper bound B.

      @param N_factors  The factors of N = p1^e1 * ... * pn^en, represented on
                        the form [[p1, e1], ..., [pn, en]], for p1, ..., pn
                        pairwise distinct prime factors, and for e1, ..., en
                        positive integer exponents.

      @return   True if (pi - 1) / gcd(pi - 1, r) is B-smooth for all but at
                most one of the prime factors pi of N, False otherwise. """

  B = floor(B);

  count = 0;

  for pi in [pi for [pi, _] in N_factors]:
    # Compute the missing factor for each prime.
    di = (pi - 1) // gcd(pi - 1, r);

    # Check if the missing factor is B-smooth. If not, add to count.
    if not is_B_smooth(di, B):
      count += 1;

  return count < 2;


def test_solve_r_for_factors_of_N(
  N : int | mpz,
  N_factors : list[list[int | mpz]],
  c : int = 1,
  verbose : bool = True) -> None:

  """ @brief  Samples the order r of an element g selected uniformly at random
              from the multiplicative group of the ring of integers modulo N,
              and attempts to recover the complete factorization of N given r.

      More specifically, this function accepts as input the complete
      factorization of an integer N = p1^e1 * ... * pn^en with n distinct prime
      factors pi, and optionally of pi - 1, for i in [1, n].

      It then samples the order r of an element g selected uniformly at random
      from the multiplicative group of the ring of integers modulo N, and
      attempts to solve r for the complete factorization of N.

      @param N  The integer N.

      @param N_factors  The factors of N = p1^e1 * ... * pn^en, represented on
                        the form [[p1, e1], ..., [pn, en]], for p1, ..., pn
                        pairwise distinct prime factors, and for e1, ..., en
                        positive integer exponents.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing smooth component in lambda'(N) when solving r for the
                complete factorization of N.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to True.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  # Sample r.
  while True:
    r = sample_r_given_N(N, N_factors);
    if r != 1:
      break;

  if verbose:
    print("Sampled r = " + str(r) + "\n");

  # Test r.
  if not is_admissible_r(r, c * N.bit_length(), N_factors):
    if verbose:
      print("\n*** Sampled r which does not admit factorization:\n", d);

    return;

  if verbose:
    print("Done setting up the problem instance. Solving commences...\n");

  # Setup and start a timer.
  timer = Timer().start();

  # Attempt to recover the factors.
  candidate_factors = solve_r_for_factors(
                        r = r,
                        N = N,
                        c = c,
                        verbose = verbose);

  # Stop the timer.
  timer.stop();

  # Compare.
  if candidate_factors != set([pi for [pi, _] in N_factors]):
    raise Exception("Error: An incorrect complete factorization was returned.");

  if verbose:
    print("\n*** Successfully recovered the complete factorization.");


def test_solve_r_for_factors_of_random_N(
  l : int = 1024,
  n : int = 2,
  e_max : int = 1,
  c : int = 1,
  verbose : bool = True) -> None:

  """ @brief  Samples the order r of an element g selected uniformly at random
              from the multiplicative group of the ring of integers modulo N,
              and attempts to recover the complete factorization of N given r.

      More specifically, this function first selects n pairwise distinct l-bit
      prime factors pi from the set of all l-bit prime factors, and n integer
      exponents ei uniformly at random from the interval [1, e_max), for i on
      [1, n], after which it computes the integer N = p1^e1 * ... * pn^en.

      It then samples the order r of an element g selected uniformly at random
      from the multiplicative group of the ring of integers modulo N, and
      attempts to solve r for the complete factorization of N.

      @remark   This convenience function simply sets up the factors of N, and
                then calls the test_solve_r_for_factors_of_N() function.

      @param l  The length in bits of each distinct prime factors of N.

      @param n  The number of distinct prime factors of N.

      @param e_max  An upper bound on the exponent of each distinct prime
                    factor.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing smooth component in lambda'(N) when solving r for the
                complete factorization of N.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to True.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  # Select factors.
  if verbose:
    print("Sampling N, please wait, this may take a moment...\n");

  factors = [];
  pis = set();

  for i in range(n):
    if e_max == 1:
      ei = 1;
    else:
      ei = int(1 + sample_integer(e_max - 1));

    while True:
      pi = mpz(2 ** (l - 1) + sample_integer(2 ** (l - 1)));

      if not pi.is_prime():
        continue;

      if pi in pis:
        continue;

      break;

    if verbose:
      print("Selected factor " + str(i) + ":", str(pi) + "^" + str(ei));

    factors.append([pi, ei]);

  # Compute N.
  N = prod([pi ** ei for [pi, ei] in factors]);

  if verbose:
    print("\nSelected N = " + str(N) + "\n");

  # Attempt to solve.
  test_solve_r_for_factors_of_N(
    N = N,
    N_factors = factors,
    c = c,
    verbose = verbose);


def test_solve_j_for_factors_of_N(
  N : int | mpz,
  N_factors : list[list[int | mpz]],
  pi_minus_one_factors : list[list[list[int | mpz]]] | None = None,
  c_factor : int = 1,
  c_solve : int = 1,
  B_solve : int = B_DEFAULT_SOLVE,
  B_sample : int = B_DEFAULT_SAMPLE,
  sample_g : bool = False,
  verbose : bool = True) -> None:

  """ @brief  Samples a frequency j from the probability distribution induced
              by Shor's order-finding algorithm when factoring N, and then
              attempts to solve j for the complete factorization of N.

      More specifically, this function accepts as input the complete
      factorization of an integer N = p1^e1 * ... * pn^en with n distinct prime
      factors pi, and optionally of pi - 1, for i in [1, n].

      It samples the order r of an element g selected uniformly at random from
      the multiplicative group of the ring of integers modulo N.

      Next, it samples a frequency j from the probability distribution induced
      by the quantum order-finding algorithm when performing order-finding with
      respect to an element of order r.

      Finally, it attempts to solve j for the complete factorization of N, by
      first solving j for r, and by then solving r for the complete
      factorization of N.

      @param N  The integer N.

      @param N_factors  The factors of N = p1^e1 * ... * pn^en, represented on
                        the form [[p1, e1], ..., [pn, en]], for p1, ..., pn
                        pairwise distinct prime factors, and for e1, ..., en
                        positive integer exponents.

      @param pi_minus_one_factors  The factors of

                                      pi - 1 = qi1^di1 * ... * qim^dim,

                                   for i in [1, n], represented on the form
                                   [F1, ..., Fn], where each Fi is on the form
                                   [[qi1, qi1], ..., [qim, qim]], for
                                   qi1, ..., qim pairwise distinct prime
                                   factors, and for di1, ..., dim positive
                                   integer exponents.

                                   May be set to None, in which case r will be
                                   computed deterministically as described
                                   above. If explicitly specified, the order r
                                   will be computed exactly.

      @param c_factor   A parameter c_factor >= 1 that specifies the maximum
                        size of the missing smooth component in lambda'(N) when
                        solving r for the complete factorization of N.

      @param c_solve  A parameter c_solve >= 1 that specifies the maximum size
                      of the missing smooth component d in r = d * r_tilde when
                      solving j for r, or a multiple of r.

      @param B_solve  The bound B on the offset in j when solving j for r.

      @param B_sample   The bound B on the offset in j when sampling j from the
                        frequency distribution induced by Shor's algorithm.

      @param sample_g   A flag that may be set to True to not exactly sample the
                        order r of an element g selected uniformly at random
                        from the multiplicative group of the ring of integers
                        modulo N, but rather to sample g and to then exactly or
                        heuristically compute r (depending on whether the
                        complete factorizations of pi - 1 for i in [1, n] are
                        specified or not).

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to True.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  m = l = N.bit_length() - 1;
  while (l > 1) and (((N / 2) ** 2) < (2 ** (m + l - 1))):
    l -= 1;

  if sample_g:
    # Sample r and g.
    if verbose:
      print("Sampling g, please wait, this may take a moment...\n");

    while True:
      [g, r] = sample_g_r_given_N(N, N_factors, pi_minus_one_factors);
      if r != 1:
        break;

    if verbose:
      print("Sampled g = " + str(g) + "\n");
      print("Heuristically computed r = " + str(r) + "\n");

    # Setup g.
    g = IntegerModRingMulSubgroupElement(g, N);
  else:
    if pi_minus_one_factors != None:
      raise Exception(
        "Error: Specify pi_minus_one_factors only when sample_g = True.");

    # Sample r.
    while True:
      r = sample_r_given_N(N, N_factors);
      if r != 1:
        break;

    if verbose:
      print("Sampled r = " + str(r) + "\n");

    # Setup g.
    g = SimulatedCyclicGroupElement(r);

  # Test r.
  if not is_admissible_r(r, c_factor * N.bit_length(), N_factors):
    if verbose:
      print("\n*** Sampled r which does not admit factorization:\n", d);

    return;

  # Sample j.
  j = sample_j_given_r(r = r, m = m, l = l, B = B_sample);
  if None == j:
    if verbose:
      print("\n*** Failed to sample j.\n");

    return;

  if verbose:
    print("Sampled j = " + str(j) + "\n");

    print("Done setting up the problem instance. Solving commences...\n");

  # Setup and start a timer.
  timer = Timer().start();

  # Attempt to recover the factors.
  candidate_factors = solve_j_for_factors(
                        j = j,
                        m = m,
                        l = l,
                        g = g,
                        N = N,
                        c_factor = c_factor,
                        c_solve = c_solve,
                        B = B_solve,
                        accept_multiple = True,
                        verbose = verbose,
                        opt_speculative = True);

  # Stop the timer.
  timer.stop();

  # Compare.
  if candidate_factors != set([pi for [pi, _] in N_factors]):
    raise Exception("Error: An incorrect complete factorization was returned.");

  if verbose:
    print("\n*** Successfully recovered the complete factorization.");


def test_solve_j_for_factors_of_random_N(
  l : int = 1024,
  n : int = 2,
  e_max : int = 1,
  c_factor : int = 1,
  c_solve : int = 1,
  B_solve : int = B_DEFAULT_SOLVE,
  B_sample : int = B_DEFAULT_SAMPLE,
  sample_g : bool = False,
  verbose : bool = True) -> None:

  """ @brief  Samples a frequency j from the probability distribution induced
              by Shor's order-finding algorithm when factoring N, and then
              attempts to solve j for the complete factorization of N.

      More specifically, this function first selects n pairwise distinct l-bit
      prime factors pi from the set of all l-bit prime factors, and n integer
      exponents ei uniformly at random from the interval [1, e_max), for i on
      [1, n], after which it computes the integer N = p1^e1 * ... * pn^en.

      It then samples the order r of an element g selected uniformly at random
      from the multiplicative group of the ring of integers modulo N. Next, it
      samples a frequency j from the probability distribution induced by the
      quantum order-finding algorithm when performing order-finding with respect
      to an element of order r.

      Finally, it attempts to solve j for the complete factorization of N, by
      first solving j for r, and by then solving r for the complete
      factorization of N.

      @remark   This convenience function simply sets up the factors of N, and
                then calls the test_solve_j_for_factors_of_N() function.

      @param l  The length in bits of each distinct prime factors of N.

      @param n  The number of distinct prime factors of N.

      @param e_max  An upper bound on the exponent of each distinct prime
                    factor.

      @param c_factor   A parameter c_factor >= 1 that specifies the maximum
                        size of the missing smooth component in lambda'(N) when
                        solving r for the complete factorization of N.

      @param c_solve  A parameter c_solve >= 1 that specifies the maximum size
                      of the missing smooth component d in r = d * r_tilde when
                      solving j for r, or a multiple of r.

      @param B_solve  The bound B on the offset in j when solving j for r.

      @param B_sample   The bound B on the offset in j when sampling j from the
                        frequency distribution induced by the quantum
                        order-finding algorithm.

      @param sample_g   A flag that may be set to True to not exactly sample the
                        order r of an element g selected uniformly at random
                        from the multiplicative group of the ring of integers
                        modulo N, but rather to sample g and to then
                        heuristically compute r.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to True.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  # Select factors.
  if verbose:
    print("Sampling N, please wait, this may take a moment...\n");

  factors = [];
  pis = set();

  for i in range(n):
    if e_max == 1:
      ei = 1;
    else:
      ei = int(1 + sample_integer(e_max - 1));

    while True:
      pi = mpz(2 ** (l - 1) + sample_integer(2 ** (l - 1)));

      if not pi.is_prime():
        continue;

      if pi in pis:
        continue;

      break;

    if verbose:
      print("Selected factor " + str(i) + ":", str(pi) + "^" + str(ei));

    factors.append([pi, ei]);

  # Compute N.
  N = prod([pi ** ei for [pi, ei] in factors]);

  if verbose:
    print("\nSelected N = " + str(N) + "\n");

  # Attempt to solve.
  test_solve_j_for_factors_of_N(
    N = N,
    N_factors = factors,
    c_factor = c_factor,
    c_solve = c_solve,
    B_solve = B_solve,
    B_sample = B_sample,
    sample_g = sample_g,
    verbose = verbose);


def test_all(verbose : bool = True) -> None:

  """ @brief  Executes the test suite described in App. A.3 of [E21b].

      [E21b] Ekerå, M.: "On completely factoring any integer efficiently in a
                         single run of an order-finding algorithm".
                        Quantum Inf. Process. 20(6):205 (2021).

      @remark   This convenience function simply iterates over all combinations
                of l in [256, 512, 1024], n in [2, 5, 10, 25], and e_max in
                [1, 2, 3], calling both test_solve_r_for_factors_of_random_N
                and test_solve_j_for_factors_of_random_N() for each combination.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to True.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  # Setup and start a timer.
  timer = Timer().start();

  for l in [256, 512, 1024]:
    for n in [2, 5, 10, 25]:
      for e_max in [1, 2, 3]:

        if verbose:
          print("\n");

        print("*** Running test for l =", str(l) + ", n =", str(n) +
          ", e_max =", str(e_max) + "...");

        # Solve the chain r -> N.
        test_solve_r_for_factors_of_random_N(
          l = l,
          n = n,
          e_max = e_max,
          verbose = verbose);

        # Solve the whole chain j -> r -> N. This goes slightly beyond [E21b].
        test_solve_j_for_factors_of_random_N(
          l = l,
          n = n,
          e_max = e_max,
          verbose = verbose);

  # Stop the timer.
  timer.stop();

  # All tests have been executed.
  print("\n*** Time required to setup and execute all tests:", timer);



""" @brief  A module for exactly or heuristically sampling an element g
            uniformly at random from the multiplicative group of the ring of
            integers modulo N and returning the order r of g. This when the
            complete factorization of N is known.

    The procedures in this module are described in [E21b], in [E24t] (see in
    particular Sect. 5.2.3) and in the Factoritall repository (available at
    https://github.com/ekera/factoritall).

    [E21b] Ekerå, M.: "On completely factoring any integer efficiently in a
                       single run of an order-finding algorithm".
                      Quantum Inf. Process. 20(6):205 (2021).

    [E24t] Ekerå, M.: "On factoring integers, and computing discrete logarithms
                       and orders, quantumly".
                      PhD thesis, KTH Royal Institute of Technology (2024). """

from gmpy2 import mpz;
from gmpy2 import powmod;

from gmpy2 import gcd;
from gmpy2 import lcm;

from math import prod;
from math import floor;

from ..math.crt import crt;
from ..math.primes import prime_range;
from ..math.random import sample_integer;

B_DEFAULT_SAMPLE = 10 ** 6;

def sample_r_given_N(
  N : int | mpz,
  factors : list[list[int | mpz]]) -> int:

  """ @brief  Returns the order r of an element g selected uniformly at random
              from the multiplicative group of the ring of integers modulo N,
              without explicitly computing and returning the element g.

      Suppose that N = p1^e1 * ... * pn^en, for p1, ..., pn pairwise distinct
      odd prime factors, and e1, ..., en positive integer exponents.

      For i in [1, n], suppose that gi is selected uniformly at random from the
      multiplicative group of the ring of integers modulo pi^ei, and that the
      order ri of gi is computed. Then r = lcm(r1, ..., rn) is the order of g,
      where g may be computed via the Chinese remainder theorem, by using that
      it must hold that gi = g mod pi^ei for i in [1, n].

      (Note that the above map is an isomorphism: Hence g will be selected
      uniformly at random from the multiplicative group of the ring of integers
      modulo N, as the gi are selected uniformly at random from the
      multiplicative group of the ring of integers modulo pi^ei.)

      A problem with the above approach is that it is hard to compute the order
      ri, and hence to compute r, unless the factorization of pi - 1 for i in
      [1, n] is known. To circumvent this problem, instead of directly selecting
      the gi uniformly at random from the multiplicative group of the ring of
      integers modulo pi^ei, suppose exponents di are instead selected uniformly
      at random on the interval [0, lambda(pi^ei)), where

          lambda(pi^ei) = (pi - 1) pi^(ei - 1)

      as all pi are odd, and gi = Gi^di computed, for Gi some fixed generator of
      the multiplicative group the ring of integers modulo pi^ei. Then, gi is of
      order ri = lambda(pi^ei) / gcd(lambda(pi^ei), di) for i in [1, n], so the
      ri are easy to compute, as is r = lcm(r1, ..., rn).

      Again, unless the factorization of pi - 1 for i in [1, n] is known, it is
      hard to prove that an element Gi is a generator, and hence to explicitly
      compute g. This explains why this function only returns r.

      The above procedure is not described in [E21b], but in [E24t] (see in
      particular Sect. 5.2.3) and in the Factoritall repository (available at
      https://github.com/ekera/factoritall).

      [E21b] Ekerå, M.: "On completely factoring any integer efficiently in a
                         single run of an order-finding algorithm".
                        Quantum Inf. Process. 20(6):205 (2021).

      [E24t] Ekerå, M.: "On factoring integers, and computing discrete
                         logarithms and orders, quantumly". PhD thesis, KTH
                        Royal Institute of Technology (2024).

      @param N  The integer N.

      @param factors  The factors of N = p1^e1 * ... * pn^en, represented on the
                      form [[p1, e1], ..., [pn, en]], for p1, ..., pn pairwise
                      distinct prime factors, and for e1, ..., en positive
                      integer exponents.

      @return   The order r of an element g selected uniformly at random from
                the multiplicative group of the ring of integers modulo N. """

  # Make sure the factors are GMP integers.
  factors = [[mpz(pi), int(ei)] for [pi, ei] in factors];
  n = len(factors);

  # Sanity checks.
  if N != prod([pi ** ei for [pi, ei] in factors]):
    raise Exception("Error: Incorrect factorization: N != sum(pi ** ei)");

  pis = [pi for [pi, _] in factors];
  if len(set(pis)) != n:
    raise Exception("Error: Incorrect factorization: " +
      "The pis are not all pairwise distinct.");

  if min(pis) <= 2:
    raise Exception("Error: Incorrect factorization: " +
      "It is required that all pi > 2.");

  for pi in pis:
    if not pi.is_prime():
      raise Exception("Error: Incorrect factorization: " +
        "It is required that all pis be prime.");

  eis = [ei for [_, ei] in factors];
  if min(eis) < 1:
    raise Exception("Error: Incorrect factorization: " +
      "It is required that all ei >= 1.");

  # Compute lambdai for i in [1, n].
  lambdais = [(pi - 1) * (pi ** (ei - 1)) for [pi, ei] in factors];

  # Sample di for i in [1, n].
  dis = [sample_integer(lambdai) for lambdai in lambdais];

  # Compute ri for i in [1, n].
  ris = [lambdais[i] // gcd(lambdais[i], dis[i]) for i in range(n)];

  # Compute r.
  r = ris[0];
  for i in range(1, n):
    r = lcm(r, ris[i]);

  # Return r.
  return int(r);


def sample_g_r_given_N(
  N : int | mpz,
  N_factors : list[list[int | mpz]],
  pi_minus_one_factors : list[list[list[int | mpz]]] | None = None,
  B : int = B_DEFAULT_SAMPLE) -> list[int]:

  """ @brief  Returns [g, r], for g an element selected uniformly at random
              from the multiplicative group of the ring of integers modulo N,
              and r either a heuristic estimate of the order of g, or the exact
              order of g, depending on if optional parameters are specified.

      Suppose that N = p1^e1 * ... * pn^en, for p1, ..., pn pairwise distinct
      odd prime factors, and e1, ..., en positive integer exponents.

      For i in [1, n], this function then selects gi from the multiplicative
      group of the ring of integers modulo pi^ei.

      1. If the factorization of pi - 1 for i in [1, n] is *not* specified, this
      function then heuristically estimates the order ri of gi by using the
      method in App. A of [E21b]: Specificially, by using that

         lambda(pi^ei) = (pi - 1) pi^(ei - 1),

      as pi is odd, and by using a factor base of primes <= B to find all small
      factors of pi - 1 via trial division. It then computes g via the Chinese
      remainder theorem, by requiring that gi = g mod pi^ei, along with a
      heuristic estimate r = lcm(r1, ..., rn) of the order of g, that is
      correct with high probability, as is explained in [E21b].

      2. If the factorization of pi - 1 for i in [1, n] is specified, this
      function then exactly computes the order ri of gi. Specifically, by using

          lambda(pi^ei) = (pi - 1) pi^(ei - 1)

      as an initial guess ri' for the order ri of gi. Then, for each prime
      factor f that divide ri', for as long as f divides ri' and
      gi^(ri' / f) = 1 (mod N), let ri' <- ri' / f. It follows that ri = ri' at
      the end of the procedure. The order of g is then r = lcm(r1, ..., rn).

      The above procedure is described in [E21b], and in the factoritall
      repository (available at https://github.com/ekera/factoritall).

      [E21b] Ekerå, M.: "On completely factoring any integer efficiently in a
                         single run of an order-finding algorithm".
                         Quantum Inf. Process. 20(6):205 (2021).

      @param N  The integer N.

      @param N_factors  The factors of N = p1^e1 * ... * pn^en, represented on
                        the form [[p1, e1], ..., [pn, en]], for p1, ..., pn
                        pairwise distinct prime factors, and for e1, ..., en
                        positive integer exponents.

      @param pi_minus_one_factors  The factors of pi-1 = qi1^di1 * ... * qim^dim,
                                   for i in [1, n], represented on the form
                                   [F1, ..., Fn], where each Fi is on the form
                                   [[qi1, qi1], ..., [qim, qim]], for
                                   qi1, ..., qim pairwise distinct prime factors,
                                   and for di1, ..., dim positive integer
                                   exponents. May be set to None, in which case
                                   r will be computed deterministically as
                                   described above. If explicitly specified, the
                                   order r will be computed exactly.

      @param B  The upper bound on the prime factors to consider when performing
                trial division. Has no effect if pi_minus_one_factors is
                explicitly specified, as trial division is then not performed.

      @return   The pair [g, r], for g an element selected uniformly at random
                from the multiplicative group of the ring of integers modulo N,
                and r a heuristic estimate of the order of g, or the exact order
                of g, depending on if optional parameters are specified. """

  # Process the factors [pi, ei] of N:

  # Make sure the factors are GMP integers.
  factors = [[mpz(pi), int(ei)] for [pi, ei] in N_factors];

  n = len(factors);

  # Sanity checks.
  if N != prod([pi ** ei for [pi, ei] in factors]):
    raise Exception("Error: Incorrect factorization of N: N != sum(pi ** ei)");

  pis = [pi for [pi, _] in factors];
  if len(set(pis)) != n:
    raise Exception("Error: Incorrect factorization of N: " +
      "The pis are not all pairwise distinct.");

  if min(pis) <= 2:
    raise Exception("Error: Incorrect factorization of N: " +
      "It is required that all pi > 2.");

  for pi in pis:
    if not pi.is_prime():
      raise Exception("Error: Incorrect factorization of N: " +
        "It is required that all pis be prime.");

  eis = [ei for [_, ei] in factors];
  if min(eis) < 1:
    raise Exception("Error: Incorrect factorization of N: " +
      "It is required that all ei >= 1.");

  if pi_minus_one_factors != None:
    # Also process the factors [qj, dj] of pi - 1:

    if len(pi_minus_one_factors) != n:
      raise Exception("Error: Incorrect factorization of pi - 1: "
        "Expected one entry for each of the n prime factors pi of N.");

    for i in range(n):
      # Make sure the factors are GMP integers.
      pi_minus_one_factors[i] = \
        [[mpz(qj), dj] for [qj, dj] in pi_minus_one_factors[i]];

      mi = len(pi_minus_one_factors[i]);

      # Sanity checks.
      if pis[i] - 1 != prod([qj ** dj for [qj, dj] in pi_minus_one_factors[i]]):
        raise Exception("Error: Incorrect factorization of pi - 1: "\
          "p" + str(i) + " - 1 != sum(qj ** dj)");

      qjs = [qj for [qj, _] in pi_minus_one_factors[i]];
      if len(set(qjs)) != mi:
        raise Exception("Error: Incorrect factorization of pi - 1: " +
          "The qjs are not all pairwise distinct.");

      if min(qjs) < 2:
        raise Exception("Error: Incorrect factorization of pi - 1: " +
          "It is required that all qj >= 2.");

      for qj in qjs:
        if not qj.is_prime():
          raise Exception("Error: Incorrect factorization of pi - 1: " +
            "It is required that all qjs be prime.");

      djs = [dj for [_, dj] in pi_minus_one_factors[i]];
      if min(djs) < 1:
        raise Exception("Error: Incorrect factorization of pi - 1: " +
          "It is required that all dj >= 1.");

  # Pre-compute the set of all primes on [2, B).
  primes = prime_range(B);

  # Compute modulii = pi^ei for i in [1, n].
  moduliis = [pi ** ei for [pi, ei] in factors];

  # Heuristically compute gi and ri for i in [1, n].
  gis = [];
  ris = [];

  for i in range(n):
    # Setup pi, ei and modulii = (pi - 1) pi^(ei - 1).
    pi = pis[i];
    ei = eis[i];
    modulii = moduliis[i];

    # Sample gi uniformly at random from Z_{pi^ei}^*.
    while True:
      gi = sample_integer(modulii);
      if gcd(gi, pi) == 1:
        break;

    # Store gi.
    gis.append(gi);

    # Branch on whether the factorization of pi - 1 is known.
    if pi_minus_one_factors != None:
      # If the factorization of pi - 1 is known, then we know the factorization
      # of the order phi(pi^ei) of Z_{pi^ei}^* and so we can exactly find the
      # order ri of the element gi sampled from Z_{pi^ei}^* above.

      # Let F be the factorization of phi(pi^ei). Since we know the complete
      # factorization of pi - 1 we can guarantee that F will be complete.
      F = pi_minus_one_factors[i];
      if ei > 1:
        F.append([pi, ei - 1]);

      # Initially let ri = phi(pi^ei) = (pi - 1) * pi^(ei - 1) / prod(F) = 1.
      ri = 1;

      # Initially let gip = gi^ri mod pi^ei = gi.
      gip = gi;
    else:
      # Let F be the factorization of phi(pi^ei). Since we do not know the
      # complete factorization of pi - 1 it may be that F will not be complete.
      F = [];
      if ei > 1:
        F.append([pi, ei - 1]);

      # Find all factors of pi - 1 that are < B and add them to F.
      tmp = pi - 1;

      for q in primes:
        d = 0;

        while (tmp % q) == 0:
          d += 1;
          tmp //= q;

        if d > 0:
          F.append([q, d]);

      # Initially let ri = phi(pi^ei) = (pi - 1) * pi^(ei - 1) / prod(F).
      ri = tmp;

      # Initially let gip = gi^ri mod pi^ei.
      gip = powmod(gi, ri, modulii);

    # The idea is now to use F to find the order of gip and to multiply it onto
    # ri. If F is complete, this will set ri to the order of gi. Otherwise, this
    # will set ri to the order of gi with high probability assuming that B is
    # large. (If ri is not set to the order of gi, ri is instead set to some
    # positive integer multiple of the order of gi.)

    # To start off, define a recursive helper function that upon input of an
    # element x of Z_{pi^ei}^* and F efficiently returns a set of tuples of the
    # form (x^(prod(F) / q^d), q, d) where q runs over the factors in F, and
    # where d is the exponent of q in F.

    def recursive(x, F):
      l = len(F);

      if l == 0:
        return set();

      if l == 1:
        [q, d] = F[0];
        return {(x, q, d)};

      F_L = F[:floor(l/2)]; F_R = F[floor(l/2):];

      d_L = mpz(prod([q ** d for [q, d] in F_R]));
      d_R = mpz(prod([q ** d for [q, d] in F_L]));

      x_L = powmod(x, d_L, modulii); x_R = powmod(x, d_R, modulii);

      return recursive(x_L, F_L).union(recursive(x_R, F_R));

    # Call the above recursive function for x = gip and F, use the set of tuples
    # (x^(prod(F) / q^d), q, d) = (gip^(phi(pi^ei) / ri / q^d), q, d) it returns
    # to identify the power of q in the order of gip, and multiply it onto ri.

    for (x, q, d) in recursive(gip, F):
      for _ in range(d):
        if x == 1:
          break;

        x = powmod(x, q, modulii);
        ri *= q;

      if x != 1:
        raise Exception("Error: Sanity check failed: Internal error.");

    # Store ri.
    ris.append(ri);

  # Compute r.
  r = ris[0];
  for i in range(1, n):
    r = lcm(r, ris[i]);

  # Compute g.
  g = crt(gis, moduliis);

  # Sanity checks.
  for i in range(n):
    if gis[i] != (g % moduliis[i]):
      raise Exception("Error: Sanity check failed: Internal error.");

  if powmod(g, r, N) != 1:
    raise Exception("Error: Sanity check failed: Internal error.");

  for q in set(primes + pis):
    if (r % q) == 0:
      if powmod(g, r // q, N) == 1:
        raise Exception("Error: Sanity check failed: Internal error.");

  # Return [g, r].
  return [int(g), int(r)];


""" @brief  A module for computing delta-LLL-reduced bases. """

import gmpy2;

from gmpy2 import mpz;
from gmpy2 import mpq;
from gmpy2 import mpfr;

from .gram_schmidt import gram_schmidt;

from ..matrices import dimensions;
from ..vectors import norm2;

from ...utils.timeout import Timeout;

LLL_DEFAULT_DELTA = 0.99;

LLL_DEFAULT_REDUCED_PRECISION = 128;

def lll(
  A : list[list[int | mpz]],
  delta : float = LLL_DEFAULT_DELTA,
  timeout : int | None | Timeout = None,
  gs : bool = False,
  precision : int | None = None) -> list[list[int | mpz]]:

  """ @brief  Returns the delta-LLL-reduced basis for an n x d basis matrix A,
              where LLL is short for Lenstra–Lenstra–Lovász [LLL82].

      [LLL82] A. K. Lenstra, H. W. Lenstra Jr. and L. Lovász. Factoring
              polynomials with rational coefficients. Math. Ann. 26(4),
              pp. 515–534 (1982).

      The basis matrix A is represented as a list of lists [a_1, ..., a_n],
      where a_i = [a_i1, ..., a_id] for i = 1, ..., n are the n rows of A. It is
      required that A has integer entries of type int or mpz.

      @param A  The basis matrix A.

      @param delta  The delta parameter in the Lovász condition. Must be on
                    the interval (1/4, 1]. A polynomial runtime in the
                    dimension n of the lattice is only guaranteed for delta < 1.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param gs  A flag that may be set to True to return not only the
                 delta-LLL-reduced basis B of A, but also the Gram–Schmidt
                 orthogonalization Bs of B, and the matrix M of Gram–Schmidt
                 projection factors such that B = M Bs. Note that Bs and M are
                 always computed as a part of the LLL reduction process.

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors in M. May be set to None, as is the
                        default, in which case the projection factors are
                        represented as exact quotients.

      @return   The delta-LLL-reduced basis B = [b_1, ..., b_n], where
                b_i = [b_i1, ..., b_id] for i = 1, ..., n represent the n row
                vectors that make up the basis, if gs is set to False.

                Otherwise, if gs is set to True, [B, [Bs, M]] is returned where
                B is as above, Bs is the Gram–Schmidt orthogonalization of B and
                M is the matrix of Gram–Schmidt projection factors. It holds
                that B = M Bs. The matrices Bs and M are represented in analogy
                with how A and B are represented. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  (n, d) = dimensions(A);

  B = [[mpz(x) for x in row] for row in A]; # = (b_1, ..., b_n)^T

  Bs = [[mpz(0) for _ in range(d)] for _ in range(n)]; # = (b_1^*, ..., b_n^*)^T

  M = [[mpz(0) for _ in range(d)] for _ in range(n)];
  for i in range(n):
    M[i][i] = mpz(1);

  with gmpy2.context(gmpy2.get_context()) as context:

    # Set the precision.
    if precision != None:
      context.precision = precision;

    def compute_mu(i, j):
      a = sum([B[i][k]  * Bs[j][k] for k in range(d)]);
      b = sum([Bs[j][k] * Bs[j][k] for k in range(d)]);

      if None == precision:
        return mpq(a, b);

      with gmpy2.context(gmpy2.get_context()) as context:

        # Set the precision.
        context.precision = precision;

        return mpfr(a) / mpfr(b);

    # 1
    Bs[0] = B[0].copy(); # b_1^* <- b_1.

    i = 2;

    # 2
    while i <= n:

      # Check the timeout.
      timeout.check();

      # 2.1: Perform size reduction for b_i and b_i^*:

      # 2.1.1
      Bs[i - 1] = [0 for _ in range(d)];

      # 2.1.2
      for j in range(i - 1, 0, -1):

        # Check the timeout.
        timeout.check();

        # 2.1.2.1
        M[i - 1][j - 1] = compute_mu(i - 1, j - 1);

        # 2.1.2.2
        if abs(M[i - 1][j - 1]) > 1 / 2:

          tmp = round(M[i - 1][j - 1]);

          # 2.1.2.2.1
          for k in range(d):
            B[i - 1][k] = B[i - 1][k] - tmp * B[j - 1][k];

          # 2.1.2.2.2
          M[i - 1][j - 1] = M[i - 1][j - 1] - tmp;

        # 2.1.2.3
        for k in range(d):
          Bs[i - 1][k] = Bs[i - 1][k] - M[i - 1][j - 1] * Bs[j - 1][k];

      # 2.1.3
      for k in range(d):
        Bs[i - 1][k] = Bs[i - 1][k] + B[i - 1][k];


      # 2.2: Enforce the Lovász condition for b_{i−1}^* and b_i^*:

      # 2.2.1
      if (delta - M[i - 1][i - 2] * M[i - 1][i - 2]) * norm2(Bs[i - 2]) <= \
        norm2(Bs[i - 1]):

        # 2.2.1.1
        i = i + 1;

      # 2.2.2
      else:

        # 2.2.2.1
        tmp = B[i - 2];
        B[i - 2] = B[i - 1];
        B[i - 1] = tmp;

        # 2.2.2.2
        if i == 2:

          # 2.2.2.2.1
          Bs[0] = B[0].copy();

        # 2.2.2.3
        else:

          # 2.2.2.3.1
          i = i - 1;

    if gs:
      return [B, [Bs, M]];

    return B;

def is_lll_reduced(
  B : list[list[int | mpz]],
  delta : float = LLL_DEFAULT_DELTA,
  gs : list[list[list[int | mpz | mpq | mpfr]]] | None = None,
  precision : int | None = None) -> bool:

  """ @brief  Checks if an n x d basis matrix B is delta-LLL-reduced, where LLL
              is short for Lenstra–Lenstra–Lovász [LLL82].

      [LLL82] A. K. Lenstra, H. W. Lenstra Jr. and L. Lovász. Factoring
              polynomials with rational coefficients. Math. Ann. 26(4),
              pp. 515–534 (1982).

      The basis matrix B is represented as a list of lists [b_1, ..., b_n],
      where b_i = [b_i1, ..., b_id] for i = 1, ..., n are the n rows of A. It is
      required that B has integer entries of type int or mpz.

      @param B    The basis matrix B.

      @param delta  The delta parameter in the Lovász condition. Must be on
                    the interval (1/4, 1].

      @param gs   The list [Bs, M], where Bs is the Gram–Schmidt orthogonalized
                  basis for B, and M is the associated matrix of Gram–Schmidt
                  projection factors, as returned by calling

                     [Bs, M] = gram_schmidt(B, precision = precision),

                  or None, in which case this function will make the above call.

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors in M. May be set to None, in which
                        case projection factors are represented as exact
                        quotients.

                        Note that this parameter only has an effect if gs is
                        set to None as the Gram–Schmidt orthogonalized basis Bs
                        and the associated matrix M of Gram–Schmidt projection
                        factors are otherwise pre-computed.

      @return   True if the basis B is delta-LLL-reduced, False otherwise. """

  # Initial setup.
  (n, _) = dimensions(B);

  if None == gs:
    [Bs, M] = gram_schmidt(B, precision = precision);
  else:
    [Bs, M] = gs;

  # Step 1: Check that B is size reduced.
  for i in range(2, n + 1):
    for j in range(1, i):
      if abs(M[i - 1][j - 1]) > 1 / 2:
        return False;

  # Step 2: Check the Lovász condition.
  for i in range(1, n - 1):
    if delta * norm2(Bs[i - 1]) > \
         norm2(Bs[i]) + M[i][i - 1] ** 2 * norm2(Bs[i - 1]):
      return False;

  # Both conditions are met so the basis B is delta-LLL-reduced.
  return True;


""" @brief  A module for testing the functions in the parent module for
            solving a frequency j yielded by the quantum part of
            Shor's order-finding algorithm for the order r, or optionally for a
            positive integer multiple of r.

    To run all unit tests in this module by calling test_all(), execute:

    ```sh

    $ python3 -m quaspy.orderfinding.general.postprocessing.ekera.test

    ``` """

from ......math.primes import is_B_smooth
from ......utils.timer import Timer;

from ......math.groups import SimulatedCyclicGroupElement;

from ......math.random import sample_integer, sample_l_bit_integer;

from ....sampling import optimal_j_for_z_r, sample_j_given_r;

from ....sampling import B_DEFAULT_SAMPLE;

from ...ekera import B_DEFAULT_SOLVE;

from .. import solve_j_for_r;

from .. import SolutionMethods;

from gmpy2 import mpz;

from math import gcd;

from ..internal.solve import solve_j_for_r_tilde_lattice_enumerate;
from ..internal.solve import solve_j_for_r_tilde_lattice_svp;
from ..internal.solve import solve_j_for_r_tilde_continued_fractions;

def test_solve_j_for_r(
  r : int | mpz,
  m : int | None = None,
  Delta : int | None = None,
  c : int = 1,
  B_solve : int = B_DEFAULT_SOLVE,
  B_sample : int = B_DEFAULT_SAMPLE,
  accept_multiple : bool = False,
  method : SolutionMethods = SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR,
  verbose : bool = True,
  opt_speculative : bool = True,
  opt_isolate_peak : bool = True) -> None:

  """ @brief  Tests the solve_j_for_r() function for a given r, and a given
              set of parameters and optimization flags.

      This function first calls sample_j_given_r(r, m, l, B = B_sample) to
      sample a Fourier frequency j from the distribution induced by the quantum
      order-finding algorithm, for r as passed to this function.

      Unless a larger value of m is explicitly specified, m is set to the bit
      length of the order r. If some Delta in [0, m) is specified, l = m-Delta,
      otherwise l is set to the least positive integer such that r^2 < 2^(m+l).

      @remark   If the sampling fails, this function returns silently, as it is
                expected that sampling j may sometimes fail.

      This function then calls

        solve_j_for_r(
          j,
          m,
          l,
          g,
          c,
          B = B_solve,
          accept_multiple,
          method,
          verbose,
          opt_speculative,
          opt_isolate_peak)

      to attempt to solve j for r, or for some positive multiple of r, depending
      on how the acccept_multiple flag is selected. Above, g is a simulated
      cyclic group element of order r (see the SimulatedCyclicGroupElement
      class), and B = B_solve. All other parameters are simply passed along.

      For the full documentation of these parameters, see solve_j_for_r().

      @param r  The order r.

      @param m  A positive integer m such that r < 2^m, or None, in which case
                this function will set m to the bit length of r.

      @param Delta  An integer in [0, m) used to select l = m-Delta, or None,
                    in which case this function will set l to the least positive
                    integer such that r^2 < 2^(m+l).

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde when solving j
                for r.

      @param B_solve  A bound B >= 0 on the offset in j when solving j for r.

                      If B > 0, this function tries to solve not only j, but
                      also j ± 1, ..., j ± B, for r, or for a positive integer
                      multiple of r. For further details, see solve_j_for_r().

      @param B_sample   A bound B >= 0 on the offset in j when sampling j. For
                        further details, see sample_j_given_r().


      @param accept_multiple  A flag that may be set to True to indicate that
                              only a positive integer multiple of r is sought.
                              If set to True, this function returns as soon as
                              it finds r such that g^r = 1.

      @param method   An enumeration entry from the SolutionMethods class that
                      specifies the method to use to solve j for r. For further
                      details, see the documentation for the SolutionMethods
                      class.

      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when executing the test.

      @param opt_speculative  A flag that may be set to True to indicate that
                              Alg. 2 in [E24] should be used instead of Alg. 3
                              to find the missing cm-smooth component of r. In
                              most cases, Alg. 2 is faster than Alg. 3, but in
                              the worst case Alg. 2 is a lot slower than Alg. 3.
                              For further details, see [E24].

      @param opt_isolate_peak   A flag that may be set to True to indicate that
                                all offsets in j up to B should not be tested
                                exhaustively. Rather, the peak around the
                                optimal frequency j_0(z) should be isolated: As
                                soon as offsets to the left and right of j_0(z)
                                are found for which the post-processing
                                algorithm fails to produce r such that g^r = 1,
                                this function returns the minimum r found such
                                that g^r = 1. Note that this flag has no effect
                                if the accept_multiple flag is set to True, as
                                this function then returns as soon as it finds r
                                such that g^r = 1.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised.

                The test succeeds if the solve_j_for_r() function behaves as
                expected, i.e. if it succeeds in all cases where it is
                guaranteed to succeed by the analysis in [E24], and either
                succeeds or fails otherwise.

                If the solve_j_for_r() function fails in a case where it is
                guaranteed to succeed in theory, an exception is raised. """

  # Select m and l = m.
  if None == m:
    m = r.bit_length();
  else:
    if r.bit_length() > m:
      raise Exception("Error: It is required that r < 2^m.");

  if None == Delta:
    l = m;
    while (l > 0) and ((r ** 2) < (2 ** (m + l - 1))):
      l -= 1;
  else:
    if (Delta < 0) or (Delta >= m):
      raise Exception("Error: It is required that Delta be on [0, m).");

    l = m - Delta;

  # Sample j.
  sample = sample_j_given_r(r, m, l, B = B_sample, extended_result = True);
  if None == sample:
    if verbose:
      print("\n*** Failed to sample j.\n");

    return;
  else:
    [j, [z, j0, j0_offset]] = sample;

  if verbose:
    print("Sampled j = " + str(j) + "\n");

  # Setup g.
  g = SimulatedCyclicGroupElement(r);

  # Setup and start a timer.
  timer = Timer().start();

  # Attempt to recover r.
  candidate_r = \
    solve_j_for_r(
      j = j,
      m = m,
      l = l,
      g = g,
      c = c,
      B = B_solve,
      accept_multiple = accept_multiple,
      method = method,
      verbose = verbose,
      opt_speculative = opt_speculative,
      opt_isolate_peak = opt_isolate_peak);

  # Stop the timer.
  timer.stop();

  if None == candidate_r:
    # Check if we expect to fail since the offset is too large.
    if abs(j0_offset) > B_solve:
      if verbose:
        print("\n*** Failed to recover r, as expected, " +
          "since the offset in j is greater than B.\n");

      return;

    # Check if we expect to fail since d = gcd(r, z) is not cm-smooth.
    d = gcd(z, r);

    if not is_B_smooth(d, c * m):
      if verbose:
        print("\n*** Failed to recover r, as expected, " +
          "since d = gcd(r, z) is not cm-smooth.\n");

      return;

    # Unexpected failure.
    if verbose:
      print("\n*** Failed to recover r.\n");

    raise Exception("Error: Failed to solve for r, despite the problem " +
      "instance being solvable in theory.");

  # Compare.
  if candidate_r == r:
    if verbose:
      print("\n*** Successfully recovered the order r.");
      print("\nTotal time required to solve j for r: " + str(timer) + "\n");

    return;

  if accept_multiple and (candidate_r >= 1) and ((candidate_r % r) == 0):
    if verbose:
      print("\n*** Successfully recovered a multiple of the order r.");
      print("\nTotal time required to solve j for a multiple r: " + \
        str(timer) + "\n");

    return;

  if verbose:
    print("\n*** Failed to recover r.\n");

  raise Exception("Error: Failed to solve for r, despite the problem " +
    "instance being solvable in theory.");


def test_solve_j_for_random_r(
  m : int = 2048,
  Delta : int | None = None,
  B_solve : int = B_DEFAULT_SOLVE,
  B_sample : int = B_DEFAULT_SAMPLE,
  accept_multiple : bool = False,
  method : SolutionMethods = SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR,
  verbose : bool = True,
  opt_speculative : bool = True,
  opt_isolate_peak : bool = True) -> None:

  """ @brief  Selects an m bit order r uniformly at random from the set of all
              such orders, and then calls test_solve_j_for_r() for this order r,
              passing along the other arguments.

      @remark   This is a convenience function. For details on all other
                parameters, please see the documentation of the
                test_solve_j_for_r() function.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  # Select an m-bit order r.
  r = sample_l_bit_integer(m);

  if verbose:
    print("Sampled r = " + str(r) + "\n");

  # Attempt to solve.
  test_solve_j_for_r(
    r = r,
    m = m,
    Delta = Delta,
    B_solve = B_solve,
    B_sample = B_sample,
    accept_multiple = accept_multiple,
    method = method,
    verbose = verbose,
    opt_speculative = opt_speculative,
    opt_isolate_peak = opt_isolate_peak);


def test_all_solve_for_r(verbose : bool = True) -> None:

  """ @brief  Calls test_solve_j_for_random_r() for all combinations of
              accept_multiple and opt_speculative in {True, False}, and, for
              each such combination, for all supported solution methods, and for
              each such combination, for all values of m in
              {128, 256, 384, 512, 1024, 2048, 4096, 8192}, passing along the
              verbose flag.

      More specifically, this function calls test_solve_j_for_random_r() for all
      combinations of accept_multiple and opt_speculative in {True, False}, and,
      for each such combination, for all solution methods in

        {SolutionMethods.CONTINUED_FRACTIONS_BASED,
         SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR;
         SolutionMethods.LATTICE_BASED_ENUMERATE},

      and, for each such combination, for all values of m in

        {128, 256, 384, 512, 1024, 2048, 4096, 8192}.

      If test_solve_j_for_random_r() raises an exception due to a test failing,
      this exception is not caught but simply allowed to propagate.

      This function always sets opt_isolate_peak to True, as is the default.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to False.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  for accept_multiple in [True, False]:
    for method in [SolutionMethods.CONTINUED_FRACTIONS_BASED,
                   SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR,
                   SolutionMethods.LATTICE_BASED_ENUMERATE]:

      if not verbose:
        print("");

      for opt_speculative in [True, False]:
        for opt_isolate_peak in [True]:

          for m in [128, 256, 384, 512, 1024, 2048, 4096, 8192]:

            if verbose:
              print("");

            if method == SolutionMethods.LATTICE_BASED_ENUMERATE:
              Deltas = [0, 1, 2, 3, 4, 5, 8, 12, 16];
            else:
              Deltas = [None];

            for Delta in Deltas:
              print("*** Running test for " +
                "m =", str(m) + ", " +
                "Delta =", str(Delta) + ", " +
                "accept_multiple =", str(accept_multiple) + ", " +
                "method =", str(method) + ", " +
                "opt_speculative =", str(opt_speculative) + ", " +
                "opt_isolate_peak =", str(opt_isolate_peak) + "...");

              test_solve_j_for_random_r(
                m = m,
                Delta = Delta,
                accept_multiple = accept_multiple,
                method = method,
                verbose = verbose,
                opt_speculative = opt_speculative,
                opt_isolate_peak = opt_isolate_peak);


def test_solve_for_r_tilde(
  z : int | mpz,
  r : int | mpz,
  m : int,
  c : int = 1,
  Deltas : bool = False,
  verbose : bool = False) -> None:

  """ @brief  Calls solve_j_for_r_tilde_continued_fractions(),
              solve_j_for_r_tilde_lattice_svp() and
              solve_j_for_r_tilde_lattice_enumerate() with j = j0(z) to test
              that r_tilde = r / gcd(r, z) is successfully recovered.

      This function picks the least positive integer l such that r^2 < 2^(m+l),
      computes the optimal frequency j = j0(z) given r, m and l, and attempts
      to solve j for r using

        - the continued fractions-based solver, by calling the
          solve_j_for_r_tilde_continued_fractions() function,

        - the lattice-based non-zero shortest vector problem (SVP) solver, by
          calling the solve_j_for_r_tilde_lattice_svp() function, and

        - the lattice-based solver with enumeration, by calling the
          solve_j_for_r_tilde_lattice_enumerate() function.

      For each solver, it is verified that r_tilde = r / gcd(r, z) is
      successfully recovered.

      If the Deltas flag is set to true, this function also attempt to solve j
      for r using the lattice-based solver with enumeration, whilst setting
      l = m - Delta och re-computing j = j0(z) given r, m and l, for all Delta
      in [0, ..., min(m - 1, 12)]. For each value of Delta, it is again verified
      that r_tilde is successfully  recovered.

      @remark   This function checks whether d = gcd(r, z) is cm-smooth.

                If d is not cm-smooth, all tests of the lattice-based solver
                with support for enumerating the lattice are skipped over the
                solver in question is not guaranteed to find r_tilde unless d is
                cm-smooth.

      @param z  An integer z on [0, r) identifying the peak index.

      @param r  The order r.

      @param m  A positive integer m such that r < 2^m.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde.

      @param Deltas   A flag that may be set to True to perform extended tests
                      for the lattice-based solver with enumeration, as
                      explained above. Defaults to False.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to False.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised."""

  l = m;
  while (l > 1) and ((r ** 2) < (2 ** (m + l - 1))):
    l -= 1;

  j0 = optimal_j_for_z_r(z = z, r = r, m = m, l = l);

  d = gcd(r, z);

  smooth = is_B_smooth(d, c * m);

  r_tilde = r // d;

  # Continued fractions:
  r_tilde_candidates = solve_j_for_r_tilde_continued_fractions(j0, m, l);

  if r_tilde != r_tilde_candidates:
    raise Exception(
      "Error: Failed to recover r_tilde by continued fractions.");

  # Lattice SVP:
  [r_tilde_candidates, _] = solve_j_for_r_tilde_lattice_svp(j0, m, l);

  if r_tilde != r_tilde_candidates:
    raise Exception(
      "Error: Failed to recover r_tilde by solving the SVP in L.");

  # Lattice enumeration:
  if smooth:
    g = SimulatedCyclicGroupElement(r);

    [r_tilde_candidates, _] = solve_j_for_r_tilde_lattice_enumerate(
                                j = j0,
                                m = m,
                                l = l,
                                g = g,
                                c = c,
                                accept_multiple = False,
                                filtered_r_tilde_candidates = set(),
                                verbose = verbose);

    if type(r_tilde_candidates) != type(set()):
      raise Exception("Error: Expected a set to the returned.");

    if not (r_tilde in r_tilde_candidates):
      raise Exception("Error: Failed to recover r_tilde by enumerating L.");


    [r_tilde_candidates, _] = solve_j_for_r_tilde_lattice_enumerate(
                                j = j0,
                                m = m,
                                l = l,
                                g = g,
                                c = c,
                                accept_multiple = False,
                                verbose = verbose);

    if not (r_tilde in r_tilde_candidates):
      raise Exception("Error: Failed to recover r_tilde by enumerating L.");


    if Deltas:

      for Delta in range(min(m - 1, 12) + 1):
        l = m - Delta;

        j0 = optimal_j_for_z_r(z = z, r = r, m = m, l = l);


        [r_tilde_candidates, _] = solve_j_for_r_tilde_lattice_enumerate(
                                    j = j0,
                                    m = m,
                                    l = l,
                                    g = g,
                                    c = c,
                                    accept_multiple = False,
                                    filtered_r_tilde_candidates = set(),
                                    verbose = verbose);

        if type(r_tilde_candidates) != type(set()):
          raise Exception("Error: Expected a set to the returned.");

        if not (r_tilde in r_tilde_candidates):
          raise Exception("Error: Failed to recover r_tilde by enumerating L.");


        [r_tilde_candidates, _] = solve_j_for_r_tilde_lattice_enumerate(
                                    j = j0,
                                    m = m,
                                    l = l,
                                    g = g,
                                    c = c,
                                    accept_multiple = False,
                                    verbose = verbose);

        if not (r_tilde in r_tilde_candidates):
          raise Exception("Error: Failed to recover r_tilde by enumerating L.");


def test_solve_for_r_tilde_exhaustive(
  m : int = 16,
  c : int = 1,
  verbose : bool = False) -> None:

  """ @brief  Selects r uniformly at random from the set of all m-bit integers,
              and then calls test_solve_for_r_tilde() for r, and for all values
              of z in [0, r), with Deltas = False, passing along the verbose
              flag.

      @param m  A positive integer m.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to False.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  # Pick an m-bit order r at random.
  r = sample_l_bit_integer(m);

  print("*** Running r_tilde recovery test for r = " + str(r) + "...\n");

  g = SimulatedCyclicGroupElement(r);

  for z in range(r):
    if verbose:
      print(" >> Processing " + str(z) + " / " + str(r));

    test_solve_for_r_tilde(
      z = z,
      r = r,
      m = m,
      c = c,
      Deltas = False,
      verbose = verbose);


def test_solve_for_r_tilde_randomized(
  m : int = 2048,
  c : int = 1,
  verbose : bool = False) -> None:

  """ @brief  Selects r uniformly at random from the set of all m-bit integers,
              then selects z uniformly at random from [0, r), and finally calls
              test_solve_for_r_tilde() for r and z, with Deltas = True, passing
              along the other arguments.

      @param m  A positive integer m.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to False.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  # Pick an m-bit order r at random.
  r = sample_l_bit_integer(m);

  print("*** Running r_tilde recovery test for r = " + str(r) + "...\n");

  z = sample_integer(r);

  test_solve_for_r_tilde(
    z = z,
    r = r,
    m = m,
    c = c,
    Deltas = True,
    verbose = verbose);


def test_all_solve_for_r_tilde(
  verbose : bool = False) -> None:

  """ @brief  Calls test_solve_for_r_tilde_exhaustive(m, verbose) for all
              integers m in {4, 5, ..., 16}, and then calls
              test_solve_for_r_tilde_randomized(m, verbose) ten times
              for m in {256, 512, 1024, 2048, 4096, 8192, 16384}, passing
              along the verbose flag.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to False.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  for m in range(4, 16 + 1):
    test_solve_for_r_tilde_exhaustive(m = m, verbose = verbose);

  for m in [256, 512, 1024, 2048, 4096, 8192, 16384]:
    for _ in range(10):
      test_solve_for_r_tilde_randomized(m = m, verbose = verbose);


def test_all(
  verbose : bool = False) -> None:

  """ @brief  Calls test_all_solve_for_r_tilde(), and then
              test_all_solve_for_r(), passing along the verbose flag.

      @param verbose  A flag that may be set to True to print intermediary
                      results. Defaults to False.

      @return   This function has no return value. If the test fails, or if some
                other error occurs, an exception is instead raised. """

  # Setup and start a timer.
  timer = Timer().start();

  test_all_solve_for_r_tilde(verbose = verbose);

  test_all_solve_for_r(verbose = verbose);

  # Stop the timer.
  timer.stop();

  if not verbose:
    print("");

  print("*** Time required to setup and execute all tests:", timer);



""" @brief  A module for solving a frequency j yielded by the quantum part of
            Shor's order-finding algorithm [Shor94] for the order r, or
            optionally for a positive integer multiple of r, by using
            the classical post-processing algorithms from [E24].

    This module furthermore contains functions for solving a list of n
    frequencies [j_1, ..., j_n] yielded by n runs of the quantum part of
    Seifert's variation [Seifert01] of Shor's order-finding algorithm [Shor94]
    for r, or for a positive integer multiple of r. This by using the classical
    post-processing algorithms from [E24t] (see Sect. 5.4) and [E21], with
    supporting functions from [E24].

    Throughout this module, the algorithms are as described in [E24t], [E24] and
    [E21]. The notation is also inherited from said works.

    [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                         In: Proceedings from FOCS '94, pp. 124–134 (1994).

    [Seifert01] Seifert, J.-P.: "Using fewer qubits in Shor's factorization
                                 algorithm via simultaneous Diophantine
                                 approximation". In: CT-RSA 2001.
                                Springer LNCS 2020, pp. 319–227 (2001).

    [E21] Ekerå, M.: "Quantum algorithms for computing general discrete
                      logarithms and orders with tradeoffs".
                     J. Math. Cryptol. 15(1), pp. 359–407 (2021).

    [E24] Ekerå, M.: "On the success probability of quantum order finding".
                     ACM Trans. Quantum Comput. 5(2):11 (2024).

    [E24t] Ekerå, M.: "On factoring integers, and computing discrete logarithms
                       and orders, quantumly". PhD thesis, KTH Royal Institute
                      of Technology (2024). """


from enum import Enum;

from gmpy2 import mpz;

from gmpy2 import gcd;

from .....math.primes import prime_power_product;

from .....math.groups import CyclicGroupElement;
from .....math.groups import IntegerModRingMulSubgroupElement;

from .internal.solve import solve_j_for_r_tilde_continued_fractions;
from .internal.solve import solve_j_for_r_tilde_lattice_svp;
from .internal.solve import solve_j_for_r_tilde_lattice_enumerate;

from .internal.collection import CandidateCollection;

from .internal.algorithms import algorithm2;
from .internal.algorithms import algorithm3;

from .....math.lattices.enumerate import enumerate as enumerate_lattice;

from .....math.lattices.lll import lll;
from .....math.lattices.lll import LLL_DEFAULT_DELTA;
from .....math.lattices.lll import LLL_DEFAULT_REDUCED_PRECISION;

from .....math.lattices.svp import solve_svp;

from .....math.vectors import norm2;

from .....utils.timeout import Timeout;

B_DEFAULT_SOLVE = 1000;


class SolutionMethods(Enum):

  """ @brief  An enumeration of methods for solving j for r.

      As is explained in [E24], these methods are all designed to solve an
      optimal frequency j = j0(z), where z in [0, r), for r_tilde = r / d,
      where d = gcd(r, z) and d is cm-smooth. In what follows below, it is
      assumed that these requirements are met.

      There are three solution methods:

      1. CONTINUED_FRACTIONS_BASED

         Expand j / 2^(m + l) in continued fractions to find z / r, and hence
         r_tilde = r / gcd(r, z), as originally proposed in [Shor94], but with
         slightly smaller m+l, as decribed in [E24].

         By Lem. 4.1 in [E24], the last convergent p / q with denominator
         q < 2^((m+l)/2) in the continued fractions expansion of j / 2^(m+l) is
         equal to z / r, provided that r < 2^m and r^2 < 2^(m+l).

         For further details, see Lem. 4.1, and Sect. 4 and App. B, of [E24].

      2. LATTICE_BASED_SHORTEST_VECTOR

         Use Lagrange's lattice basis reduction algorithm to find the shortest
         non-zero vector in the lattice L generated by (j, 1/2) and
         (2^(m+l), 0).

         By Lem. 4.2 in [E24], provided that r < 2^m and r^2 < 2^(m+l), the
         second component of the shortest non-zero vector in L has r_tilde / 2
         in its second component, up to sign of course.

         For further details, see Lem. 4.2, and Sect. 4 and App. C, of [E24].

      3. LATTICE_BASED_ENUMERATE

         Use Lagrange's lattice basis reduction algorithm to find a reduced
         basis for the lattice L spanned by (j, 1/2) and (2^(m+l), 0), and
         enumerate all vectors within a ball of radius 2^(m-1/2) in L centered
         at the origin to find u = (rj - 2^(m+l) z, r / 2) / d, and hence
         r_tilde, as the second component is r_tilde / 2.

         By Lem. 4.3 in [E24], provided that r < 2^m and l = m - Delta, at most
         6 * sqrt(3) * 2^Delta vectors must be enumerated in L to find u and
         hence r_tilde, so if Delta is small then this method is efficient.

         In practice, as mentioned in [E24], the leading constant in the above
         bound is not tight, and the enumeration can be optimized. Some of these
         optimizations are implemented here so the enumeration typically
         considers fewer vectors than the bound indicates.

         For further details, see Lem. 4.3, and Sect. 4 and App. C, of [E24].

      [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                           In: Proceedings from FOCS '94, pp. 124–134 (1994).

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024). """

  CONTINUED_FRACTIONS_BASED = 1;

  LATTICE_BASED_SHORTEST_VECTOR = 2;

  LATTICE_BASED_ENUMERATE = 3;


class EnumerationOptions(Enum):

  """ @brief  An enumeration of options for solving a list of n frequencies
              [j_1, ..., j_n] for r.

      The meaning of the options are explained in the documentation for the
      solve_multiple_j_for_r() and solve_multiple_j_for_r_mod_N() functions. """

  TRUE = True;

  FALSE = False;

  SVP = "SVP";

  BOUNDED_BY_TAU = "BOUNDED_BY_TAU";

  BOUNDED_BY_TAU_COMPLETE = "BOUNDED_BY_TAU_COMPLETE";


def solve_j_for_r(
  j : int | mpz,
  m : int,
  l : int,
  g : CyclicGroupElement,
  c : int = 1,
  B : int = B_DEFAULT_SOLVE,
  accept_multiple : bool = False,
  method : SolutionMethods = SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR,
  timeout : int | None | Timeout = None,
  verbose : bool  = False,
  opt_speculative : bool = True,
  opt_isolate_peak : bool = True):

  """ @brief  Attempts to compute the order r of g, or a positive integer
              multiple thereof, given a frequency j yielded by the quantum part
              of Shor's order-finding algorithm [Shor94], by using the
              post-processing algorithms described in detail in [E24].

      [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                           In: Proceedings from FOCS '94, pp. 124–134 (1994).

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      The idea is to try to solve not only j, but also j ± 1, ..., j ± B, for r,
      with the aim of solving an optimal frequency j0(z) for r, for z the peak
      index on [0, r). Provided

        - that j0(z) is solved for r,

        - that d = gcd(r, z) is cm-smooth
          (for the definition of cm-smooth in [E24]),

        - that l is selected as required by the solution method
          (see the documentation for parameter l below), and

        - that the accept_multiple flag is set to False,

      the order r will be found by this function.

      Note that this function does not implement meet-in-the-middle-techniques,
      although it is noted in [E23p] that it is possible to use such techniques
      to speed up the post-processing also for order finding.

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025).

      @param j  The frequency j yielded by the quantum order-finding algorithm.

      @param m  A positive integer m such that r < 2^m.

      @param l  A positive integer l <= m, such that m + l is the length of the
                control register in the quantum order-finding algorithm.

                If method is set to SolutionMethods.CONTINUED_FRACTIONS_BASED or
                SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR, it is required
                that r^2 < 2^(m + l) or else r may not be found.

                If method is set to SolutionMethods.LATTICE_BASED_ENUMERATE, it
                is possible to select l = m - Delta for some Delta in [0, m), at
                the expense of enumerating at most 6 * sqrt(3) * 2^Delta lattice
                vectors for each offset in j considered.

      @param g  The group element g of order r.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde when solving j
                for r, for cm-smooth as defined in [E24].

                As is explained in [E24], increasing c increases the success
                probability, at the expense of increasing the runtime.

      @param B  A bound B >= 0 on the offset in j. If B > 0, this function tries
                to solve not only j, but also j ± 1, ..., j ± B, for r, or for a
                positive integer multiple of r.

      @param accept_multiple  A flag that may be set to True to indicate that
                              only a positive integer multiple of r is sought.
                              If set to True, this function returns as soon as
                              it finds r such that g^r = 1.

      @param method   An enumeration entry from the SolutionMethods class that
                      specifies the method to use to solve j for r. For further
                      details, see the documentation for the SolutionMethods
                      class.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when executing the
                      post-processing algorithm.

      @param opt_speculative  A flag that may be set to True to indicate that
                              Alg. 2 in [E24] should be used instead of Alg. 3
                              to find the missing cm-smooth component of r. In
                              most cases, Alg. 2 is faster than Alg. 3, but in
                              the worst case Alg. 2 is a lot slower than Alg. 3.
                              For further details, see [E24].

      @param opt_isolate_peak   A flag that may be set to True to indicate that
                                all offsets in j up to B should not be tested
                                exhaustively. Rather, the peak around the
                                optimal frequency j_0(z) should be isolated: As
                                soon as offsets to the left and right of j_0(z)
                                are found for which the post-processing
                                algorithm fails to produce r such that g^r = 1,
                                this function returns the minimum r found such
                                that g^r = 1. Note that this flag has no effect
                                if the accept_multiple flag is set to True, as
                                this function then returns as soon as it finds r
                                such that g^r = 1.

      @return   If the accept_multiple flag is set to False, the order r of g
                is returned with probability >= P, for P as given by the lower
                bound in [E24] (provided that the opt_isolate_peak flag is set
                to False). Otherwise, None, or exceptionally a positive integer
                multiple of the order r, is returned. If the accept_multiple
                flag is set to True, some positive integer multiple of r is
                returned with probabilty >= P. Otherwise, None is returned. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  # Configure algorithms.
  if opt_speculative:
    algorithm = algorithm2;
  else:
    algorithm = algorithm3;

  # Setup the set S', and keep it reduced by using a special class.
  filtered_r_tilde_candidates = CandidateCollection();

  # Setup a set of dismissed reduced candidates for r_tilde.
  dismissed_reduced_r_tilde_candidates = set();

  # Initially set mu to zero. For each r_tilde_candidate to be tested, check if
  #
  #   gp^(reduced_r_tilde_candidate) = g^(reduced_r_tilde_candidate * e) = 1
  #
  # for reduced_r_tilde_candidate = gcd(r_tilde_candidate, mu). If so, we know
  # that reduced_r_tilde_candidate * e is a multiple of r, so we can update mu
  # by mu <- gcd(mu, reduced_r_tilde_candidate * e). This also explains why for
  # each r_tilde_candidate, it is sufficient to test reduced_r_tilde_candidate.

  mu = mpz(0);

  # Pre-compute e and gp.
  e = prime_power_product(c * m);

  gp = g ** e;
  if gp == 1:
    # Trivial case.
    return algorithm(g, 1, m, c, timeout);

  j = mpz(j);

  pow2ml = 2 ** (m + l);
  if B > pow2ml // 2:
    B = pow2ml // 2;
  pow2ml = mpz(pow2ml);

  [left, right] = [None, None];
  [skip_left, skip_right] = [False, False];

  positive_multiples = None;
  negative_multiples = None;

  for offset in range(0, B + 1):
    if verbose:
      print("Trying offset:", offset, "/", B);

    if skip_left and skip_right:
      break;

    for sign in [1, -1]:

      # Check the timeout.
      timeout.check();

      if (0 == offset) and (-1 == sign):
        continue;

      if opt_isolate_peak:
        if [left, right] != [None, None]:
          if sign == 1:
            if offset > right + 1:
              skip_right = True;
              continue;
          else:
            if -offset < left - 1:
              skip_left = True;
              continue;

      offset_j = (j + sign * offset) % pow2ml;

      # Use the prescribed solution method to find candidates for r_tilde.
      if method == SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR:

        # Fetch an initial guess for the row multiples.
        if offset == 0:
          multiples = None;
        else:
          if sign == 1:
            multiples = positive_multiples;
          else:
            multiples = negative_multiples;

        # Assume that we are not able to solve.
        success = False;

        # Solve for r_tilde.
        [r_tilde_candidate, multiples] = \
          solve_j_for_r_tilde_lattice_svp(offset_j, m, l, multiples);
        if (r_tilde_candidate >= 1) and (r_tilde_candidate < (2 ** m)):
          r_tilde_candidates = [r_tilde_candidate];
        else:
          r_tilde_candidates = [];

        # Update the guess for the row multiples.
        if offset == 0:
          positive_multiples = negative_multiples = multiples;
        else:
          if sign == 1:
            positive_multiples = multiples;
          else:
            negative_multiples = multiples;

      elif method == SolutionMethods.LATTICE_BASED_ENUMERATE:

        # Fetch an initial guess for the row multiples.
        if offset == 0:
          multiples = None;
        else:
          if sign == 1:
            multiples = positive_multiples;
          else:
            multiples = negative_multiples;

        # Solve for r_tilde.
        [filtered_r_tilde_candidates,
          [success,
           dismissed_reduced_r_tilde_candidates,
           mu,
           multiples]] = \
          solve_j_for_r_tilde_lattice_enumerate(
            j = offset_j,
            m = m,
            l = l,
            g = g,
            g_pow_e_context = [gp, e],
            c = c,
            accept_multiple = accept_multiple,
            filtered_r_tilde_candidates = filtered_r_tilde_candidates,
            dismissed_reduced_r_tilde_candidates =
              dismissed_reduced_r_tilde_candidates,
            multiples = multiples,
            verbose = verbose);

        # Update the guess for the row multiples.
        if offset == 0:
          positive_multiples = negative_multiples = multiples;
        else:
          if sign == 1:
            positive_multiples = multiples;
          else:
            negative_multiples = multiples;

      elif method == SolutionMethods.CONTINUED_FRACTIONS_BASED:

        # Assume that we are not able to solve.
        success = False;

        # Solve for r_tilde.
        r_tilde_candidate = \
          solve_j_for_r_tilde_continued_fractions(offset_j, m, l);
        if (r_tilde_candidate >= 1) and (r_tilde_candidate < (2 ** m)):
          r_tilde_candidates = [r_tilde_candidate];
        else:
          r_tilde_candidates = [];

      else:
        raise Exception("Error: Incorrect parameters: Unknown method.");


      if method in [SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR,
                    SolutionMethods.CONTINUED_FRACTIONS_BASED]:

        # When using either of these solution methods, the candidates in
        # r_tilde_candidates need not fulfill the requirement that
        #
        #   gp^r_tilde_candidate == 1,
        #
        # so we need to check this requirement. We do this by first checking
        # if r_tilde_candidate is already in the filtered set of candidates for
        # r_tilde that have passed this requirement. Only if r_tilde_candidate
        # is not in filtered_r_tilde_candidates do we need to exponentiate.

        for r_tilde_candidate in r_tilde_candidates:
          if r_tilde_candidate in filtered_r_tilde_candidates:
            # We already found this candidate, or a divisor of it.
            success = True;
          else:
            # Use that mu is an r-multiple to reduce the candidate for r_tilde.
            reduced_r_tilde_candidate = gcd(r_tilde_candidate, mu);

            # Check if the remainder after reduction is equal to one, or if we
            # already have dismissed this reduced candidate. Otherwise proceed.
            if (reduced_r_tilde_candidate == 1) or \
               (reduced_r_tilde_candidate in \
                dismissed_reduced_r_tilde_candidates):
              # Dismiss the reduced candidate.
              if verbose:
                print("Dismissing:", r_tilde_candidate);
            else:
              # The reduced candidate has not already been dismissed.
              if verbose:
                print("Testing the reduced candidate:", \
                  reduced_r_tilde_candidate);

              # Test the reduced candidate.
              if (gp ** reduced_r_tilde_candidate) == 1:
                if accept_multiple:
                  # Return immediately.
                  return algorithm(g, r_tilde_candidate, m, c, timeout);

                # Add r_tilde_candidate to the filtered candidates for r_tilde.
                filtered_r_tilde_candidates.add(r_tilde_candidate);

                # We know that reduced_r_tilde_candidate * e is a multiple of r,
                # so we may update mu to account for this fact:
                mu = gcd(reduced_r_tilde_candidate * e, mu);

                # Mark this j as successful.
                success = True;
              else:
                # Add reduced_r_tilde_candidate to the dismissed reduced
                # candidates for r_tilde to avoid future exponentiations.
                dismissed_reduced_r_tilde_candidates.\
                  add(reduced_r_tilde_candidate);

      elif method == SolutionMethods.LATTICE_BASED_ENUMERATE:

        # When using this solution method, the enumerate_lattice() function will
        # already have tested the requirement that
        #
        #   gp^r_tilde_candidate == 1,
        #
        # as very many candidates could otherwise be returned. Hence, there is
        # no reason to repeat the test here by exponentiating.

        if accept_multiple:
          if len(filtered_r_tilde_candidates) != 0:
            # Return immediately.
            return algorithm(g, min(filtered_r_tilde_candidates), m, c, timeout);

      else:
        raise Exception("Error: Incorrect parameters: Unknown method.");

      if opt_isolate_peak and success:
        if [left, right] == [None, None]:
          left = right = offset;
        elif sign == 1:
          if right == offset - 1:
            right = offset;
          else:
            raise Exception("Error: Not a contiguous interval for the peak.");
        else:
          if left == -offset + 1:
            left = -offset;
          else:
            raise Exception("Error: Not a contiguous interval for the peak.");

  if len(filtered_r_tilde_candidates) == 0:
    return None;

  return min([algorithm(g, r_tilde_candidate, m, c, timeout)
    for r_tilde_candidate in filtered_r_tilde_candidates])


def solve_j_for_r_mod_N(
  j : int | mpz,
  m : int,
  l : int,
  g : int | mpz,
  N : int | mpz,
  c : int = 1,
  B : int = B_DEFAULT_SOLVE,
  accept_multiple : bool = False,
  method : SolutionMethods = SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR,
  timeout : int | None | Timeout = None,
  verbose : bool = False,
  opt_speculative : bool = True,
  opt_isolate_peak : bool = True):

  """ @brief  Attempts to compute the order r of g mod N, or a positive integer
              multiple thereof, given a frequency j yielded by the quantum part
              of Shor's order-finding algorithm [Shor94], by using the
              post-processing algorithms described in detail in [E24].

      @remark   This convenience function simply calls solve_j_for_r() with g
                setup by calling IntegerModRingMulSubgroupElement(g, N).

      [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                           In: Proceedings from FOCS '94, pp. 124–134 (1994).

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      The idea is to try to solve not only j, but also j ± 1, ..., j ± B, for r,
      with the aim of solving an optimal frequency j0(z) for r, for z the peak
      index on [0, r). Provided

        - that j0(z) is solved for r,

        - that d = gcd(r, z) is cm-smooth
          (for the definition of cm-smooth in [E24]),

        - that l is selected as required by the solution method
          (see the documentation for parameter l below), and

        - that the accept_multiple flag is set to False,

      the order r will be found by this function.

      Note that this function does not implement meet-in-the-middle-techniques,
      although it is noted in [E23p] that it is possible to use such techniques
      to speed up the post-processing also for order finding.

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025).

      @param j  The frequency j. An integer on [0, 2^(m + l)).

      @param m  A positive integer m such that r < 2^m.

      @param l  A positive integer l <= m, such that m + l is the length of the
                control register in the quantum order-finding algorithm.

                If method is set to SolutionMethods.CONTINUED_FRACTIONS_BASED or
                SolutionMethods.LATTICE_BASED_SHORTEST_VECTOR, it is required
                that r^2 < 2^(m + l) or else r may not be found.

                If method is set to SolutionMethods.LATTICE_BASED_ENUMERATE, it
                is possible to select l = m - Delta for some Delta in [0, m), at
                the expense of enumerating at most 6 * sqrt(3) * (2 ** Delta)
                lattice vectors for each offset in j considered.

      @param g  The group element g of order r modulo N.

      @param N  The modulus N.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde when solving j
                for r, for cm-smooth as defined in [E24].

                As is explained in [E24], increasing c increases the success
                probability, at the expense of increasing the runtime.

      @param B  A bound B >= 0 on the offset in j. If B > 0, this function tries
                to solve not only j, but also j ± 1, ..., j ± B, for r, or for a
                positive integer multiple of r.

      @param accept_multiple  A flag that may be set to True to indicate that
                              only a positive integer multiple of r is sought.
                              If set to True, this function returns as soon as
                              it finds r such that g^r = 1.

      @param method   An enumeration entry from the SolutionMethods class that
                      specifies the method to use to solve j for r. For further
                      details, see the documentation for the SolutionMethods
                      class.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when executing the
                      post-processing algorithm.

      @param opt_speculative  A flag that may be set to True to indicate that
                              Alg. 2 in [E24] should be used instead of Alg. 3
                              to find the missing cm-smooth component of r. In
                              most cases, Alg. 2 is faster than Alg. 3, but in
                              the worst case Alg. 2 is a lot slower than Alg. 3.
                              For further details, see [E24].

      @param opt_isolate_peak   A flag that may be set to True to indicate that
                                all offsets in j up to B should not be tested
                                exhaustively. Rather, the peak around the
                                optimal frequency j_0(z) should be isolated: As
                                soon as offsets to the left and right of j_0(z)
                                are found for which the post-processing
                                algorithm fails to produce r such that g^r = 1,
                                this function returns the minimum r found such
                                that g^r = 1. Note that this flag has no effect
                                if the accept_multiple flag is set to True, as
                                this function then returns as soon as it finds r
                                such that g^r = 1.

      @return   If the accept_multiple flag is set to False, the order r of g
                mod N is returned with probability >= P, for P as given by the
                lower bound in [E24] (provided that the opt_isolate_peak flag is
                set to False). Otherwise, None, or exceptionally a positive
                integer multiple of the order r, is returned. If the
                accept_multiple flag is set to True, some positive integer
                multiple of r is returned with probabilty >= P. Otherwise, None
                is returned. """

  g = IntegerModRingMulSubgroupElement(g, N);

  return solve_j_for_r(
    j = j,
    m = m,
    l = l,
    g = g,
    c = c,
    B = B,
    accept_multiple = accept_multiple,
    method = method,
    timeout = timeout,
    verbose = verbose,
    opt_speculative = opt_speculative,
    opt_isolate_peak = opt_isolate_peak);


def solve_multiple_j_for_r(
  j_list : list[int | mpz],
  m : int,
  l : int,
  g : CyclicGroupElement,
  tau : int = 0,
  c : int = 1,
  delta : float = LLL_DEFAULT_DELTA,
  precision : int | None = None,
  enumerate : bool | EnumerationOptions = False,
  timeout : int | None | Timeout = None,
  verbose : bool = False,
  opt_speculative : bool = True):

  """ @brief  Attempts to compute the order r of g, or a positive integer
              multiple thereof, given a list of n frequencies [j_1, ..., j_n]
              yielded by n independent runs of the quantum part of Seifert's
              variation [Seifert01] of Shor's order-finding algorithm [Shor94]
              as described in [E24t] (see Sect. 5.4) and [E21] (see App. A).
              This by using Ekerå's lattice-based classical post-processing
              from [E24t] and [E21], with supporting functions from [E24].

      [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                           In: Proceedings from FOCS '94, pp. 124–134 (1994).

      [Seifert01] Seifert, J.-P.: "Using fewer qubits in Shor's factorization
                                   algorithm via simultaneous Diophantine
                                   approximation". In: CT-RSA 2001.
                                  Springer LNCS 2020, pp. 319–227 (2001).

      [E21] Ekerå, M.: "Quantum algorithms for computing general discrete
                        logarithms and orders with tradeoffs".
                       J. Math. Cryptol. 15(1), pp. 359–407 (2021).

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      [E24t] Ekerå, M.: "On factoring integers, and computing discrete
                         logarithms and orders, quantumly".
                        PhD thesis, KTH Royal Institute of Technology (2024).

      @param j_list   The n frequencies [j_1, ..., j_n] where j_1, ..., j_n are
                      integers on [0, 2^(m + l)).

      @param m  A positive integer m such that r < 2^m.

      @param l  A positive integer l ≈ m / s for s a tradeoff factor.

      @param g  The group element g of order r.

      @param tau  An integer tau on [0, l]. Used to scale the basis for the
                  lattice L^tau that is used in the post-processing, and that is
                  generated by the vector (j_1, ..., j_n, 2^tau), and by the n
                  vectors (2^(m + l), 0, ..., 0) thru (0, ..., 0, 2^(m + l), 0).

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde when solving j
                for r, for cm-smooth as defined in [E24].

                As is explained in [E24], increasing c increases the success
                probability, at the expense of increasing the runtime.

      @param delta  The parameter delta to use when delta-LLL-reducing the basis
                    for the lattice L^tau used in the post-processing. Must be
                    on the interval (1/4, 1]. A polynomial runtime in the
                    dimension of the lattice is only guaranteed for delta < 1.

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors as a part of delta-LLL-reducing the
                        basis for the lattice L^tau used in the post-processing.

                        The precision may be set to None, as is the default, in
                        which case the projection factors are represented as
                        exact quotients.

      @param enumerate  A flag that may be set to True to enumerate vectors in
                        the lattice L^tau (until r or a positive multiple of r
                        is found or the specified timeout has elapsed), or to
                        EnumerationOptions.SVP to consider only a shortest
                        non-zero vector in L^tau as returned by performing a
                        limited enumeration, or to False to consider only a
                        shortest non-zero vector as returned by LLL.

                        May also be set to EnumerationOptions.BOUNDED_BY_TAU or
                        to EnumerationOptions.BOUNDED_BY_TAU_COMPLETE in which
                        case all vectors within distance R of the origin of the
                        lattice L^tau are enumerated, where R depends on tau as
                        R = sqrt(n + 1) * 2^(m + tau). In the former case, the
                        enumeration is aborted early as soon as r or a positive
                        multiple of r is found. In the latter case, the
                        enumeration runs to completion, after which the minimum
                        candidate for r is taken as r.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when executing the
                      post-processing algorithm.

      @param opt_speculative  A flag that may be set to True to indicate that
                              Alg. 2 in [E24] should be used instead of Alg. 3
                              to find the missing cm-smooth component of r. In
                              most cases, Alg. 2 is faster than Alg. 3, but in
                              the worst case Alg. 2 is a lot slower than Alg. 3.
                              For further details, see [E24].

      @return   The order r of g, or some positive multiple thereof, or None, if
                solving for r fails. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  if None in j_list:
    return None;

  n = len(j_list);
  if n < 1:
    return None;

  # Configure algorithms.
  if opt_speculative:
    algorithm = algorithm2;
  else:
    algorithm = algorithm3;

  # Pre-compute e and gp.
  e = prime_power_product(c * m);

  gp = g ** e;
  if gp == 1:
    # Trivial case.
    return algorithm(g, 1, m, c, timeout);

  # Setup a basis for the lattice.
  if verbose:
    print("Setting up the basis...");

  A = [[mpz(j) for j in j_list] + [mpz(2 ** tau)]];
  for i in range(0, n):
    row = (n + 1) * [mpz(0)];
    row[i] = mpz(2 ** (m + l));
    A.append(row);

  # Reduce the basis for the lattice.
  if verbose:
    print("Computing the LLL-reduced basis with reduced precision...");

  B = lll(A,
          delta = delta,
          timeout = timeout,
          precision = LLL_DEFAULT_REDUCED_PRECISION);
            # Pre-compute with reduced precision.

  # Check the timeout.
  timeout.check();

  if verbose:
    print(" ** Refining the LLL-reduced basis with full precision...");

  [B, [Bs, M]] = lll(B,
                     delta = delta,
                     timeout = timeout,
                     gs = True,
                     precision = precision);

  # Check the timeout.
  timeout.check();


  if enumerate in [EnumerationOptions.TRUE, True,
                   EnumerationOptions.FALSE, False,
                   EnumerationOptions.SVP]:

    # Find a short non-zero vector in the lattice.
    if enumerate == EnumerationOptions.SVP:
      if verbose:
        print("Executing an enumeration to solve SVP...");

      # Let u be a shortest non-zero vector in the lattice.
      u = solve_svp(B, timeout = timeout, gs = [Bs, M]);
    else:
      # Let u be a shortest non-zero vector in the delta-LLL-reduced basis.
      u = B[0];

    # Check if a solution was found.
    if verbose:
      print(" ** Checking:", u);

    r_tilde_candidate = abs(u[-1]);
    if 0 < r_tilde_candidate < 2 ** (m + tau):
      r_tilde_candidate = r_tilde_candidate // (2 ** tau);
      if (gp ** r_tilde_candidate) == 1:
        return algorithm(g, r_tilde_candidate, m, c, timeout);

    if enumerate not in [True, EnumerationOptions.TRUE]:
      return None;

    # Check the timeout.
    timeout.check();

    if verbose:
      print("Enumerating vectors around v...");

    # Set the enumeration radius R = |u|^2.
    R2 = norm2(u);

    # Enumerate with gradually increasing radii.
    while True:

      # Check the timeout.
      timeout.check();

      # Enumerate, whilst checking the timeout.
      for u in enumerate_lattice(B,
                                 R2,
                                 timeout = timeout,
                                 gs = [Bs, M]):

        # Check if a solution was found.
        if verbose:
          print(" ** Checking:", u);

        r_tilde_candidate = abs(u[-1]);
        if 0 < r_tilde_candidate < 2 ** (m + tau):
          r_tilde_candidate = r_tilde_candidate // (2 ** tau);
          if (gp ** r_tilde_candidate) == 1:
            return algorithm(g, r_tilde_candidate, m, c, timeout);

      # Double the enumeration radius.
      if verbose:
          print(" ** Doubling the enumeration radius.");

      R2 = 4 * R2;

  elif enumerate in [EnumerationOptions.BOUNDED_BY_TAU,
                     EnumerationOptions.BOUNDED_BY_TAU_COMPLETE]:

    mu = mpz(0);

    # Setup the set S', and keep it reduced by using a special class.
    filtered_r_tilde_candidates = CandidateCollection();

    # Setup a set of dismissed reduced candidates for r_tilde.
    dismissed_reduced_r_tilde_candidates = set();

    # Check the timeout.
    timeout.check();

    # Enumerate, whilst checking the timeout.
    for u in enumerate_lattice(B,
                               R2 = mpz(n + 1) * mpz(2 ** (2 * (m + tau))),
                               timeout = timeout,
                               gs = [Bs, M]):

      # Check if a solution was found.
      if verbose:
        print(" ** Checking:", u);

      r_tilde_candidate = abs(u[-1]);
      if 0 < r_tilde_candidate < 2 ** (m + tau):
        r_tilde_candidate = r_tilde_candidate // (2 ** tau);

        if r_tilde_candidate in filtered_r_tilde_candidates:
          # We already found this candidate, or a divisor of it.
          continue;

        # Use that mu is an r-multiple to reduce the candidate for r_tilde.
        reduced_r_tilde_candidate = gcd(mu, r_tilde_candidate);

        # Check if the remainder after reduction is equal to one, or if we
        # already have dismissed this reduced candidate. Otherwise proceed.
        if ((reduced_r_tilde_candidate == 1) or
            (reduced_r_tilde_candidate in \
               dismissed_reduced_r_tilde_candidates)):
          # We can dismiss this candidate.
          continue;

        # Test the reduced candidate.
        if (gp ** reduced_r_tilde_candidate) == 1:
          if enumerate == EnumerationOptions.BOUNDED_BY_TAU:
            # Return immediately.
            return algorithm(g, r_tilde_candidate, m, c, timeout);

          # Add r_tilde_candidate to the filtered candidates for r_tilde.
          filtered_r_tilde_candidates.add(r_tilde_candidate);

          # We know that reduced_r_tilde_candidate * e is a multiple of r,
          # so we may update mu to account for this fact:
          mu = gcd(reduced_r_tilde_candidate * e, mu);
        else:
          dismissed_reduced_r_tilde_candidates.add(reduced_r_tilde_candidate);

    if len(filtered_r_tilde_candidates) == 0:
      # No solution was found.
      return None;

    return min([algorithm(g, r_tilde_candidate, m, c, timeout)
      for r_tilde_candidate in filtered_r_tilde_candidates])

  else:
    raise Exception("Error: Incorrect parameters: Unknown enumerate option.");


def solve_multiple_j_for_r_mod_N(
  j_list : list[int | mpz],
  m : int,
  l : int,
  g : int | mpz,
  N : int | mpz,
  tau : int = 0,
  c : int = 1,
  delta : float = LLL_DEFAULT_DELTA,
  precision : int | None = None,
  enumerate : bool | EnumerationOptions = False,
  timeout : int | None | Timeout = None,
  verbose : bool = False,
  opt_speculative : bool = True):

  """ @brief  Attempts to compute the order r of g mod N, or a positive integer
              multiple thereof, given a list of n frequencies [j_1, ..., j_n]
              yielded by n independent runs of the quantum part of Seifert's
              variation [Seifert01] of Shor's order-finding algorithm [Shor94]
              as described in [E24t] (see Sect. 5.4) and [E21] (see App. A).
              This by using Ekerå's lattice-based classical post-processing
              from [E24t] and [E21], with supporting functions from [E24].

      @remark   This convenience function simply calls solve_multiple_j_for_r()
                with g setup by calling IntegerModRingMulSubgroupElement(g, N).

      [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                           In: Proceedings from FOCS '94, pp. 124–134 (1994).

      [Seifert01] Seifert, J.-P.: "Using fewer qubits in Shor's factorization
                                   algorithm via simultaneous Diophantine
                                   approximation". In: CT-RSA 2001.
                                  Springer LNCS 2020, pp. 319–227 (2001).

      [E21] Ekerå, M.: "Quantum algorithms for computing general discrete
                        logarithms and orders with tradeoffs".
                       J. Math. Cryptol. 15(1), pp. 359–407 (2021).

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      [E24t] Ekerå, M.: "On factoring integers, and computing discrete
                         logarithms and orders, quantumly".
                        PhD thesis, KTH Royal Institute of Technology (2024).

      @param j_list   The n frequencies [j_1, ..., j_n] where j_1, ..., j_n are
                      integers on [0, 2^(m + l)).

      @param m  A positive integer m such that r < 2^m.

      @param l  A positive integer l ≈ m / s for s a tradeoff factor.

      @param g  The group element g of order r modulo N.

      @param N  The modulus N.

      @param tau  An integer tau on [0, l]. Used to scale the basis for the
                  lattice L^tau that is used in the post-processing, and that is
                  generated by the vector (j_1, ..., j_n, 2^tau), and by the n
                  vectors (2^(m + l), 0, ..., 0) thru (0, ..., 0, 2^(m + l), 0).

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde when solving j
                for r, for cm-smooth as defined in [E24].

                As is explained in [E24], increasing c increases the success
                probability, at the expense of increasing the runtime.

      @param delta  The parameter delta to use when delta-LLL-reducing the basis
                    for the lattice L^tau used in the post-processing. Must be
                    on the interval (1/4, 1]. A polynomial runtime in the
                    dimension of the lattice is only guaranteed for delta < 1.

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors as a part of delta-LLL-reducing the
                        basis for the lattice L^tau used in the post-processing.

                        The precision may be set to None, as is the default, in
                        which case the projection factors are represented as
                        exact quotients.

      @param enumerate  A flag that may be set to True to enumerate vectors in
                        the lattice L^tau (until r or a positive multiple of r
                        is found or the specified timeout has elapsed), or to
                        EnumerationOptions.SVP to consider only a shortest
                        non-zero vector in L^tau as returned by performing a
                        limited enumeration, or to False to consider only a
                        shortest non-zero vector as returned by LLL.

                        May also be set to EnumerationOptions.BOUNDED_BY_TAU or
                        to EnumerationOptions.BOUNDED_BY_TAU_COMPLETE in which
                        case all vectors within distance R of the origin of the
                        lattice L^tau are enumerated, where R depends on tau as
                        R = sqrt(n + 1) * 2^(m + tau). In the former case, the
                        enumeration is aborted early as soon as r or a positive
                        multiple of r is found. In the latter case, the
                        enumeration runs to completion, after which the minimum
                        candidate for r is taken as r.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when executing the
                      post-processing algorithm.

      @param opt_speculative  A flag that may be set to True to indicate that
                              Alg. 2 in [E24] should be used instead of Alg. 3
                              to find the missing cm-smooth component of r. In
                              most cases, Alg. 2 is faster than Alg. 3, but in
                              the worst case Alg. 2 is a lot slower than Alg. 3.
                              For further details, see [E24].

      @return   The order r of g mod N, or some positive multiple thereof, or
                None, if solving for r fails. """

  g = IntegerModRingMulSubgroupElement(g, N);

  return solve_multiple_j_for_r(
    j_list = j_list,
    m = m,
    l = l,
    g = g,
    tau = tau,
    c = c,
    delta = delta,
    precision = precision,
    enumerate = enumerate,
    timeout = timeout,
    verbose = verbose,
    opt_speculative = opt_speculative);



""" @brief  A module for factoring an integer N given the order r of an element
            g selected uniformly at random from the multiplicative group of the
            ring of integers modulo N. """